#!/usr/bin/env python
"""
Provide a TOKIO-aware HDF5 class that knows how to interpret schema versions
encoded in a TOKIO HDF5 file and translate a universal schema into file-specific
schemas.  Also supports dynamically mapping static HDF5 datasets into new
derived datasets dynamically.
"""

import math
import time
import datetime
import warnings
import h5py
import numpy
import pandas
import tokio.common
from tokio.connectors._hdf5 import (convert_counts_rates, #pylint: disable=unused-import
                                    map_dataset,
                                    demux_column,
                                    get_timestamps,
                                    get_timestamps_key,
                                    DEFAULT_TIMESTAMP_DATASET,
                                    TIMESTAMP_KEY,
                                    COLUMN_NAME_KEY)

SCHEMA = {
    None: {},
    "1": {
        "datatargets/readbytes": "/datatargets/readbytes",
        "datatargets/writebytes": "/datatargets/writebytes",
        "datatargets/readrates": "/datatargets/readrates",
        "datatargets/writerates": "/datatargets/writerates",
        "datatargets/readops": "/datatargets/readops",
        "datatargets/writeops": "/datatargets/writeops",
        "datatargets/readoprates": "/datatargets/readoprates",
        "datatargets/writeoprates": "/datatargets/writeoprates",
        "mdtargets/opens": "/mdtargets/opens",
        "mdtargets/openrates": "/mdtargets/openrates",
        "mdtargets/closes": "/mdtargets/closes",
        "mdtargets/closerates": "/mdtargets/closerates",
        "mdtargets/mknods": "/mdtargets/mknods",
        "mdtargets/mknodrates": "/mdtargets/mknodrates",
        "mdtargets/links": "/mdtargets/links",
        "mdtargets/linkrates": "/mdtargets/linkrates",
        "mdtargets/unlinks": "/mdtargets/unlinks",
        "mdtargets/unlinkrates": "/mdtargets/unlinkrates",
        "mdtargets/mkdirs": "/mdtargets/mkdirs",
        "mdtargets/mkdirrates": "/mdtargets/mkdirrates",
        "mdtargets/rmdirs": "/mdtargets/rmdirs",
        "mdtargets/rmdirrates": "/mdtargets/rmdirrates",
        "mdtargets/renames": "/mdtargets/renames",
        "mdtargets/renamerates": "/mdtargets/renamerates",
        "mdtargets/getxattrs": "/mdtargets/getxattrs",
        "mdtargets/getxattrrates": "/mdtargets/getxattrrates",
        "mdtargets/statfss": "/mdtargets/statfss",
        "mdtargets/statfsrates": "/mdtargets/statfsrates",
        "mdtargets/setattrs": "/mdtargets/setattrs",
        "mdtargets/setattrrates": "/mdtargets/setattrrates",
        "mdtargets/getattrs": "/mdtargets/getattrs",
        "mdtargets/getattrrates": "/mdtargets/getattrrates",
        "mdservers/cpuuser": "/mdservers/cpuuser",
        "mdservers/cpusys": "/mdservers/cpusys",
        "mdservers/cpuidle": "/mdservers/cpuidle",
        "mdservers/cpuload": "/mdservers/cpuload",
        "mdservers/memfree": "/mdservers/memfree",
        "mdservers/memused": "/mdservers/memused",
        "mdservers/memcached": "/mdservers/memcached",
        "mdservers/membuffered": "/mdservers/membuffered",
        "mdservers/memslab": "/mdservers/memslab",
        "mdservers/memslab_unrecl": "/mdservers/memslab_unrecl",
        "mdservers/memtotal": "/mdservers/memtotal",
        "dataservers/cpuuser": "/dataservers/cpuuser",
        "dataservers/cpusys": "/dataservers/cpusys",
        "dataservers/cpuidle": "/dataservers/cpuidle",
        "dataservers/cpuload": "/dataservers/cpuload",
        "dataservers/memfree": "/dataservers/memfree",
        "dataservers/memused": "/dataservers/memused",
        "dataservers/memcached": "/dataservers/memcached",
        "dataservers/membuffered": "/dataservers/membuffered",
        "dataservers/memslab": "/dataservers/memslab",
        "dataservers/memslab_unrecl": "/dataservers/memslab_unrecl",
        "dataservers/memtotal": "/dataservers/memtotal",
        "fullness/bytes": "/fullness/bytes",
        "fullness/bytestotal": "/fullness/bytestotal",
        "fullness/inodes": "/fullness/inodes",
        "fullness/inodestotal": "/fullness/inodestotal",
        "failover/datatargets": "/failover/datatargets",
        "failover/mdtargets": "/failover/mdtargets",
    },
}

# Map keys which don't exist as datasets in the underlying HDF5 but can be
# calculated from datasets that _do_ exist to the functions that do these
# conversions.  This table is only consulted when a dataset is not found
# directly in the underlying HDF5 _and_ a mapping from the SCHEMA table
# above does not return a match, so this table contains what appear to be
# some circular references (e.g., both datatargets/readbytes and
# datatargets/readrates).  This allows any valid HDF5 file to contain either
# bytes or rates but have them all present the same datasets to the downstream
# application.
SCHEMA_DATASET_PROVIDERS = {
    None: {
        "datatargets/readbytes": {
            'func': convert_counts_rates,
            'args': {
                'from_key': 'OSTReadGroup/OSTBulkReadDataSet',
                'to_rates': False,
                'transpose': True,
            },
        },
        "datatargets/writebytes": {
            'func': convert_counts_rates,
            'args': {
                'from_key': 'OSTWriteGroup/OSTBulkWriteDataSet',
                'to_rates': False,
                'transpose': True,
            },
        },
        "datatargets/readrates": {
            'func': map_dataset,
            'args': {
                'from_key': "/OSTReadGroup/OSTBulkReadDataSet",
                'transpose': True,
            },
        },
        "datatargets/writerates": {
            'func': map_dataset,
            'args': {
                'from_key': "/OSTWriteGroup/OSTBulkWriteDataSet",
                'transpose': True,
            },
        },
        "dataservers/cpuload": {
            'func': map_dataset,
            'args': {
                'from_key': "/OSSCPUGroup/OSSCPUDataSet",
                'transpose': True,
            },
        },
        "mdservers/cpuload": {
            'func': map_dataset,
            'args': {
                'from_key': "/MDSCPUGroup/MDSCPUDataSet",
                'transpose': True,
                'force2d': True,
            },
        },

        ### MDSOpsGroup, as counts per timestep
        "mdtargets/opens": {
            'func': demux_column,
            'args': {
                'from_key': "/MDSOpsGroup/MDSOpsDataSet",
                'column': 'open',
                'apply_timestep_func': lambda x, timestep: x * timestep,
                'transpose': True
            },
        },
        "mdtargets/closes": {
            'func': demux_column,
            'args': {
                'from_key': '/MDSOpsGroup/MDSOpsDataSet',
                'column': 'close',
                'apply_timestep_func': lambda x, timestep: x * timestep,
                'transpose': True,
            },
        },
        "mdtargets/mknods": {
            'func': demux_column,
            'args': {
                'from_key': '/MDSOpsGroup/MDSOpsDataSet',
                'column': 'mknod',
                'apply_timestep_func': lambda x, timestep: x * timestep,
                'transpose': True,
            },
        },
        "mdtargets/links": {
            'func': demux_column,
            'args': {
                'from_key': '/MDSOpsGroup/MDSOpsDataSet',
                'column': 'link',
                'apply_timestep_func': lambda x, timestep: x * timestep,
                'transpose': True,
            },
        },
        "mdtargets/unlinks": {
            'func': demux_column,
            'args': {
                'from_key': '/MDSOpsGroup/MDSOpsDataSet',
                'column': 'unlink',
                'apply_timestep_func': lambda x, timestep: x * timestep,
                'transpose': True,
            },
        },
        "mdtargets/mkdirs": {
            'func': demux_column,
            'args': {
                'from_key': '/MDSOpsGroup/MDSOpsDataSet',
                'column': 'mkdir',
                'apply_timestep_func': lambda x, timestep: x * timestep,
                'transpose': True,
            },
        },
        "mdtargets/rmdirs": {
            'func': demux_column,
            'args': {
                'from_key': '/MDSOpsGroup/MDSOpsDataSet',
                'column': 'rmdir',
                'apply_timestep_func': lambda x, timestep: x * timestep,
                'transpose': True,
            },
        },
        "mdtargets/renames": {
            'func': demux_column,
            'args': {
                'from_key': '/MDSOpsGroup/MDSOpsDataSet',
                'column': 'rename',
                'apply_timestep_func': lambda x, timestep: x * timestep,
                'transpose': True,
            },
        },
        "mdtargets/getxattrs": {
            'func': demux_column,
            'args': {
                'from_key': '/MDSOpsGroup/MDSOpsDataSet',
                'column': 'getxattr',
                'apply_timestep_func': lambda x, timestep: x * timestep,
                'transpose': True,
            },
        },
        "mdtargets/statfss": {
            'func': demux_column,
            'args': {
                'from_key': '/MDSOpsGroup/MDSOpsDataSet',
                'column': 'statfs',
                'apply_timestep_func': lambda x, timestep: x * timestep,
                'transpose': True,
            },
        },
        "mdtargets/setattrs": {
            'func': demux_column,
            'args': {
                'from_key': '/MDSOpsGroup/MDSOpsDataSet',
                'column': 'setattr',
                'apply_timestep_func': lambda x, timestep: x * timestep,
                'transpose': True,
            },
        },
        "mdtargets/getattrs": {
            'func': demux_column,
            'args': {
                'from_key': '/MDSOpsGroup/MDSOpsDataSet',
                'column': 'getattr',
                'apply_timestep_func': lambda x, timestep: x * timestep,
                'transpose': True,
            },
        },
        ### MDSOpsGroup, as counts per second
        "mdtargets/openrates": {
            'func': demux_column,
            'args': {
                'from_key': "/MDSOpsGroup/MDSOpsDataSet",
                'column': 'open',
                'transpose': True
            },
        },
        "mdtargets/closerates": {
            'func': demux_column,
            'args': {
                'from_key': '/MDSOpsGroup/MDSOpsDataSet',
                'column': 'close',
                'transpose': True,
            },
        },
        "mdtargets/mknodrates": {
            'func': demux_column,
            'args': {
                'from_key': '/MDSOpsGroup/MDSOpsDataSet',
                'column': 'mknod',
                'transpose': True,
            },
        },
        "mdtargets/linkrates": {
            'func': demux_column,
            'args': {
                'from_key': '/MDSOpsGroup/MDSOpsDataSet',
                'column': 'link',
                'transpose': True,
            },
        },
        "mdtargets/unlinkrates": {
            'func': demux_column,
            'args': {
                'from_key': '/MDSOpsGroup/MDSOpsDataSet',
                'column': 'unlink',
                'transpose': True,
            },
        },
        "mdtargets/mkdirrates": {
            'func': demux_column,
            'args': {
                'from_key': '/MDSOpsGroup/MDSOpsDataSet',
                'column': 'mkdir',
                'transpose': True,
            },
        },
        "mdtargets/rmdirrates": {
            'func': demux_column,
            'args': {
                'from_key': '/MDSOpsGroup/MDSOpsDataSet',
                'column': 'rmdir',
                'transpose': True,
            },
        },
        "mdtargets/renamerates": {
            'func': demux_column,
            'args': {
                'from_key': '/MDSOpsGroup/MDSOpsDataSet',
                'column': 'rename',
                'transpose': True,
            },
        },
        "mdtargets/getxattrrates": {
            'func': demux_column,
            'args': {
                'from_key': '/MDSOpsGroup/MDSOpsDataSet',
                'column': 'getxattr',
                'transpose': True,
            },
        },
        "mdtargets/statfsrates": {
            'func': demux_column,
            'args': {
                'from_key': '/MDSOpsGroup/MDSOpsDataSet',
                'column': 'statfs',
                'transpose': True,
            },
        },
        "mdtargets/setattrrates": {
            'func': demux_column,
            'args': {
                'from_key': '/MDSOpsGroup/MDSOpsDataSet',
                'column': 'setattr',
                'transpose': True,
            },
        },
        "mdtargets/getattrrates": {
            'func': demux_column,
            'args': {
                'from_key': '/MDSOpsGroup/MDSOpsDataSet',
                'column': 'getattr',
                'transpose': True,
            },
        },
    },
    "1": {
        "datatargets/readbytes": {
            'func': convert_counts_rates,
            'args': {
                'from_key': 'datatargets/readrates',
                'to_rates': False,
            },
        },
        "datatargets/writebytes": {
            'func': convert_counts_rates,
            'args': {
                'from_key': 'datatargets/writerates',
                'to_rates': False,
            },
        },
        "datatargets/readrates": {
            'func': convert_counts_rates,
            'args': {
                'from_key': 'datatargets/readbytes',
                'to_rates': True,
            },
        },
        "datatargets/writerates": {
            'func': convert_counts_rates,
            'args': {
                'from_key': 'datatargets/writebytes',
                'to_rates': True,
            },
        },
        "mdtargets/opens": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/openrates",
                'to_rates': False,
            },
        },
        "mdtargets/closes": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/closerates",
                'to_rates': False,
            },
        },
        "mdtargets/mknods": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/mknodrates",
                'to_rates': False,
            },
        },
        "mdtargets/links": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/linkrates",
                'to_rates': False,
            },
        },
        "mdtargets/unlinks": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/unlinkrates",
                'to_rates': False,
            },
        },
        "mdtargets/mkdirs": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/mkdirrates",
                'to_rates': False,
            },
        },
        "mdtargets/rmdirs": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/rmdirrates",
                'to_rates': False,
            },
        },
        "mdtargets/renames": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/renamerates",
                'to_rates': False,
            },
        },
        "mdtargets/getxattrs": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/getxattrrates",
                'to_rates': False,
            },
        },
        "mdtargets/statfss": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/statfsrates",
                'to_rates': False,
            },
        },
        "mdtargets/setattrs": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/setattrrates",
                'to_rates': False,
            },
        },
        "mdtargets/getattrs": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/getattrrates",
                'to_rates': False,
            },
        },
        "mdtargets/openrates": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/opens",
                'to_rates': True,
            },
        },
        "mdtargets/closerates": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/closes",
                'to_rates': True,
            },
        },
        "mdtargets/mknodrates": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/mknods",
                'to_rates': True
            },
        },
        "mdtargets/linkrates": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/links",
                'to_rates': True,
            },
        },
        "mdtargets/unlinkrates": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/unlinks",
                'to_rates': True,
            },
        },
        "mdtargets/mkdirrates": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/mkdirs",
                'to_rates': True,
            },
        },
        "mdtargets/rmdirrates": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/rmdirs",
                'to_rates': True,
            },
        },
        "mdtargets/renamerates": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/renames",
                'to_rates': True,
            },
        },
        "mdtargets/getxattrrates": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/getxattrs",
                'to_rates': True,
            },
        },
        "mdtargets/statfsrates": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/statfss",
                'to_rates': True,
            },
        },
        "mdtargets/setattrrates": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/setattrs",
                'to_rates': True,
            },
        },
        "mdtargets/getattrrates": {
            'func': convert_counts_rates,
            'args': {
                'from_key': "mdtargets/getattrs",
                'to_rates': True,
            },
        },
    },
}

H5LMT_COLUMN_ATTRS = {
    'MDSOpsGroup/MDSOpsDataSet': 'OpNames',
    'OSTReadGroup/OSTBulkReadDataSet': 'OSTNames',
    'OSTWriteGroup/OSTBulkWriteDataSet': 'OSTNames',
    'OSSCPUGroup/OSSCPUDataSet': 'OSSNames',
}

class Hdf5(h5py.File):
    """Hdf5 file class with extra hooks to parse different schemas

    Provides an h5py.File-like class with added methods to provide a generic
    API that can decode different schemata used to store file system load
    data.

    Attributes:
        always_translate (bool): If True, looking up datasets by keys will
            always attempt to map that key to a new dataset according to the
            schema even if the key matches the name of an existing dataset.
        dataset_providers (dict): Map of logical dataset names (keys) to dicts
            that describe the functions used to convert underlying literal
            dataset data into the format expected when dereferencing the logical
            dataset name.
        schema (dict): Map of logical dataset names (keys) to the literal
            dataset names in the underlying file (values)
        _version (str): Defined and used at initialization time to determine
            what schema to apply to map the HDF5 connector API to the underlying
            HDF5 file.
        _timesteps (dict): Keyed by dataset name (str) and has values
            corresponding to the timestep (in seconds) between each sampled
            datum in that dataset.
    """
    def __init__(self, *args, **kwargs):
        """Initialize an HDF5 file

        This is just an HDF5 file object; the magic is in the additional methods
        and indexing that are provided by the TOKIO Time Series-specific HDF5
        object.

        Args:
            ignore_version (bool): If true, do not throw KeyError if the HDF5
                file does not contain a valid version.
        """
        ignore_version = kwargs.pop('ignore_version', False)

        super(Hdf5, self).__init__(*args, **kwargs)

        # If True, always translate __getitem__ requests according to the
        # schema, even if __getitem__ requests a dataset that exists
        self.always_translate = False

        self._version = self.attrs.get('version')
        if isinstance(self._version, bytes):
            self._version = self._version.decode()
        self._timesteps = {}

        # Connect the schema map to this object
        if self._version in SCHEMA:
            self.schema = SCHEMA[self._version]
        elif self._version is None:
            self.schema = {}
        elif not ignore_version:
            raise KeyError("Unknown schema version %s" % self._version)

        # Connect the schema dataset providers to this object
        if self._version in SCHEMA_DATASET_PROVIDERS:
            self.dataset_providers = SCHEMA_DATASET_PROVIDERS[self._version]
        else:
            self.dataset_providers = {}

    def __getitem__(self, key):
        """Resolve dataset names into actual data

        Provides a single interface through which standard keys can be
        dereferenced and a semantically consistent view of data is returned
        regardless of the schema of the underlying HDF5 file.

        Passes through the underlying h5py.Dataset via direct access or a 1:1
        mapping between standardized key and an underlying dataset name, or a
        numpy array if an underlying h5py.Dataset must be transformed to match the
        structure and semantics of the data requested.

        Can also suffix datasets with special meta-dataset names
        (e.g., "/missing") to access data that is related to the root
        dataset.

        Args:
            key (str): The standard name of a dataset to be accessed.

        Returns:
            h5py.Dataset or numpy.ndarray:

              * h5py.Dataset if key is a literal dataset name
              * h5py.Dataset if key maps directly to a literal dataset name
                given the file schema version
              * numpy.ndarray if key maps to a provider function that can
                calculate the requested data
        """
        if not self.always_translate and super(Hdf5, self).__contains__(key):
            # If the dataset exists in the underlying HDF5 file, just return it
            return super(Hdf5, self).__getitem__(key)

        # Quirky way to access the missing data of a dataset through the
        # __getitem__ API via recursion
        base_key, modifier = reduce_dataset_name(key)
        if modifier and modifier == 'missing':
            return self.get_missing(base_key)

        resolved_key, provider = self._resolve_schema_key(key)

        if resolved_key:
            # Otherwise, attempt to map the logical key to a literal key
            return super(Hdf5, self).__getitem__(resolved_key)
        elif provider:
            # Or run the value through a key provider
            provider_func = provider.get('func')
            provider_args = provider.get('args', {})
            if provider_func is None:
                errmsg = "No provider function for %s" % key
                raise KeyError(errmsg)
            else:
                return provider_func(self, **provider_args)
        else:
            # This should never be hit based on the possible outputs of _resolve_schema_key
            errmsg = "_resolve_schema_key: undefined output from %s" % key
            raise KeyError(errmsg)

    def _resolve_schema_key(self, key):
        """
        Given a key, either return a key that can be used to index self
        directly, or return a provider function and arguments to generate the
        dataset dynamically
        """
        if super(Hdf5, self).__contains__(key):
            # If the dataset exists in the underlying HDF5 file, just return it
            return key, None

        # Straight mapping between the key and a dataset
        key = key.lstrip('/') if tokio.common.isstr(key) else key
        if key in self.schema:
            hdf5_key = self.schema.get(key)
            if super(Hdf5, self).__contains__(hdf5_key):
                return hdf5_key, None

        # Key maps to a transformation
        if key in self.dataset_providers:
            return None, self.dataset_providers[key]

        errmsg = "Unknown key %s in %s" % (key, self.filename)
        raise KeyError(errmsg)

    def get_version(self, dataset_name=None):
        """Get the version attribute from an HDF5 file dataset

        Args:
            dataset_name (str): Name of dataset to retrieve version.  If None,
                return the global file's version.
        Returns:
            str: The version string for the specified dataset
        """
        if dataset_name is None:
            return self._version
        else:
            # resolve dataset name
            dataset = self.__getitem__(dataset_name)
            try:
                # dataset can be either an HDF5 dataset or numpy.ndarray
                version = dataset.attrs.get("version")
            except AttributeError:
                version = None
            if version is None:
                version = self._version
            if isinstance(version, bytes):
                return version.decode() # for python3
            return version

    def set_version(self, version, dataset_name=None):
        """Set the version attribute from an HDF5 file dataset

        Provide a portable way to set the global schema version or the version
        of a specific dataset.

        Args:
            version (str): The new version to be set
            dataset_name (str): Name of dataset to set version.  If None,
                set the global file's version.
        """
        if dataset_name is None:
            self._version = version
            return self._version

        # resolve dataset name
        dataset = self.__getitem__(dataset_name)
        if dataset is None:
            raise KeyError("Dataset %s does not exist" % dataset_name)
        dataset.attrs["version"] = version
        return version

    def get_columns(self, dataset_name):
        """Get the column names of a dataset

        Args:
            dataset_name (str): name of dataset whose columns will be retrieved

        Returns:
            numpy.ndarray: Array of column names, or empty if no columns defined
        """
        # Look for special 'missing' dataset hack
        if len(dataset_name.strip('/').split('/')) == 3:
            dataset_name = dataset_name.rsplit('/', 1)[0]

        if self.get_version(dataset_name=dataset_name) is None:
            return self._get_columns_h5lmt(dataset_name)

        return self.__getitem__(dataset_name).attrs.get(COLUMN_NAME_KEY).astype('U')

    def _get_columns_h5lmt(self, dataset_name):
        """Get the column names of an h5lmt dataset
        """
        dataset = self.__getitem__(dataset_name)
        orig_dataset_name = dataset_name.lstrip('/')
        dataset_name = dataset.name.lstrip('/')
        if dataset_name == 'MDSOpsGroup/MDSOpsDataSet' and orig_dataset_name != dataset_name:
            return numpy.array([SCHEMA_DATASET_PROVIDERS[None][orig_dataset_name]['args']['column']])
        elif dataset_name in H5LMT_COLUMN_ATTRS:
            return dataset.attrs[H5LMT_COLUMN_ATTRS[dataset_name]].astype('U')
        elif dataset_name == 'MDSCPUGroup/MDSCPUDataSet':
            return numpy.array(['_unknown'])
        elif dataset_name == 'FSMissingGroup/FSMissingDataSet':
            return numpy.array(['_unknown%04d' % i for i in range(dataset.shape[1])])
        else:
            raise KeyError('Unknown h5lmt dataset %s' % dataset_name)

    def get_timestep(self, dataset_name, timestamps=None):
        """
        Cache or calculate the timestep for a dataset
        """
        if dataset_name not in self._timesteps:
            if timestamps is None:
                timestamps = self.get_timestamps(dataset_name)[0:2]
            self._timesteps[dataset_name] = timestamps[1] - timestamps[0]
        return self._timesteps[dataset_name]

    def get_index(self, dataset_name, target_datetime):
        """
        Turn a datetime object into an integer that can be used to reference
        specific times in datasets.
        """
        timestamps = self.get_timestamps(dataset_name)[0:2]
        timestep = self.get_timestep(dataset_name, timestamps)
        t_start = datetime.datetime.fromtimestamp(timestamps[0])
        return int((target_datetime - t_start).total_seconds() / timestep)

    def get_timestamps(self, dataset_name):
        """Return timestamps dataset corresponding to given dataset name

        This method returns a dataset, not a numpy array, so you can face severe
        performance penalties trying to iterate directly on the return value!
        To iterate over timestamps, it is almost always better to dereference
        the dataset to get a numpy array and iterate over that in memory.

        Args:
            dataset_name (str): Logical name of dataset whose timestamps should
                be retrieved

        Returns:
            h5py.Dataset: The dataset containing the timestamps corresponding to
            dataset_name.
        """
        return get_timestamps(self, dataset_name)

    def get_missing(self, dataset_name, inverse=False):
        """Convert a dataset into a matrix indicating the abscence of data

        Args:
            dataset_name (str): name of dataset to access
            inverse (bool): return 0 for missing and 1 for present if True

        Returns:
            numpy.ndarray: Array of numpy.int8 of 1 and 0 to indicate the
            presence or absence of specific elements
        """
        if self.get_version(dataset_name=dataset_name) is None:
            return self._get_missing_h5lmt(dataset_name, inverse=inverse)
        return missing_values(self[dataset_name][:], inverse)

    def _get_missing_h5lmt(self, dataset_name, inverse=False):
        """Return the FSMissingGroup dataset from an H5LMT file

        Encodes a hot mess of hacks to return something that looks like what
        `get_missing()` would return for a real dataset.

        Args:
            dataset_name (str): name of dataset to access
            inverse (bool): return 0 for missing and 1 for present if True

        Returns:
            numpy.ndarray: Array of numpy.int8 of 1 and 0 to indicate the
            presence or absence of specific elements
        """
        dataset = self.__getitem__(dataset_name)
        missing_dataset = self.get('/FSMissingGroup/FSMissingDataSet')
        if len(dataset.shape) == 1:
            result = numpy.zeros((dataset.shape[0], 1), dtype=numpy.int8)
        elif dataset.shape == missing_dataset.shape:
            result = missing_dataset[:, :].astype('i8').T
        else:
            result = numpy.zeros(dataset[:, :].shape, dtype=numpy.int8).T

        if inverse:
            return (~result.astype(bool)).astype('i8')
        return result

    def to_dataframe(self, dataset_name):
        """Convert a dataset into a dataframe

        Args:
            dataset_name (str): dataset name to convert to DataFrame

        Returns:
            pandas.DataFrame: DataFrame indexed by datetime objects
            corresponding to timestamps, columns labeled appropriately, and
            values from the dataset
        """
        if self.get_version(dataset_name=dataset_name) is None:
            return self._to_dataframe_h5lmt(dataset_name)
        return self._to_dataframe(dataset_name)

    def _to_dataframe(self, dataset_name):
        """Convert a dataset into a dataframe via TOKIO HDF5 schema
        """
        values = self[dataset_name][:]
        columns = self.get_columns(dataset_name)
        timestamps = self.get_timestamps(dataset_name)[...]
        if len(columns) < values.shape[1]:
            columns.resize(values.shape[1])

        # transform missing data into NaNs
        mask = missing_values(values) != 0
        try:
            values[mask] = numpy.nan
        except ValueError: # ValueError: cannot convert float NaN to integer
            # don't bother converting non-float arrays' -0.0 into NaNs
            pass

        dataframe = pandas.DataFrame(data=values,
                                     index=[datetime.datetime.fromtimestamp(t) for t in timestamps],
                                     columns=columns)
        return dataframe

    def _to_dataframe_h5lmt(self, dataset_name):
        """Convert a dataset into a dataframe via H5LMT native schema
        """
        normed_name, modifier = reduce_dataset_name(dataset_name)
        if not modifier:
            normed_name = dataset_name.lstrip('/')
        else:
            normed_name = normed_name.lstrip('/')

        col_header_key = H5LMT_COLUMN_ATTRS.get(normed_name)

        # Hack around datasets that lack column headers to retrieve column names
        if col_header_key is not None:
            columns = self[dataset_name].attrs[col_header_key]
        elif normed_name == 'FSMissingGroup/FSMissingDataSet':
            columns = self['/OSSCPUGroup/OSSCPUDataSet'].attrs['OSSNames']
        elif normed_name == 'MDSCPUGroup/MDSCPUDataSet':
            columns = ['unknown_mds']
        else:
            columns = None

        # Get timestamps through regular API
        timestamps = self.get_timestamps(normed_name)[...]

        # Retrieve and transform data using H5LMT schema directly
        if normed_name == 'FSStepsGroup/FSStepsDataSet':
            values = None
        else:
            num_dims = len(self[dataset_name].shape)
            if num_dims == 1:
                values = self[dataset_name][:]
                num_indices_expected = self[dataset_name].shape[0]
            elif num_dims == 2:
                # only transpose if dataset_name refers to a native type
                if normed_name in SCHEMA_DATASET_PROVIDERS[None]:
                    values = self[dataset_name][:]
                    columns = self.get_columns(normed_name)
                else:
                    values = self[dataset_name][:].T
                num_indices_expected = values.shape[0]
            elif num_dims > 2:
                raise Exception("Can only convert 1d or 2d datasets to dataframe")

        indices = [datetime.datetime.fromtimestamp(t) for t in timestamps]
        num_indices = len(indices)

        # if an HDF5 file was initialized but not fully populated, the number of
        # timestamps (indices) might be less than the size of the matrix being
        # stored.  this is OK as long as there are no valid values in the part
        # of the matrix that don't have corresponding indices.
        if num_indices_expected != num_indices:
            warning_msg = "dataset size and timestamps are inconsistent (%d vs %d values; missing sum is %f)" \
                % (num_indices_expected, num_indices, values[num_indices:, :].sum())
            missings = self.get_missing(normed_name)
#           print("shape of defined range: (%d, %d)" % missings[:num_indices, :].shape)
#           print("shape of undefined range: (%d, %d)" % missings[num_indices:, :].shape)
#           print("missing elements in defined range:", missings[:num_indices, :].sum())
#           print("missing elements in undefined range:", missings[num_indices:, :].sum())

            # if we have more indices than values, this is unrecoverable because
            # we don't know where in `values` data was written twice
            if num_indices > num_indices_expected:
                raise IndexError(warning_msg)

            # h5lmt format doesn't support distinguishing uninitialized elements
            # from initialized-but-zero, so we assume that a zero indicates
            # missing here
            if (self.get_version() is None and values[num_indices:, :].sum() != 0.0) \
            or (self.get_version() is not None and ~(missings[num_indices:, :].astype(bool)).sum() != 0):
                raise IndexError(warning_msg)
            if num_dims == 1:
                values = values[:num_indices]
            else:
                values = values[:num_indices, :]
            warnings.warn(warning_msg)

        return pandas.DataFrame(data=values,
                                index=indices,
                                columns=columns)

    def to_timeseries(self, dataset_name, light=False):
        """Creates a TimeSeries representation of a dataset

        Create a TimeSeries dataset object with the data from an existing HDF5
        dataset.

        Responsible for setting timeseries.dataset_name, timeseries.columns, timeseries.dataset,
        timeseries.dataset_metadata, timeseries.group_metadata, timeseries.timestamp_key

        Args:
            dataset_name (str): Name of existing dataset in self to convert into
                a TimeSeries object
            light (bool): If True, don't actually load datasets into memory;
                reference them directly into the HDF5 file

        Returns:
            tokio.timeseries.TimeSeries: The in-memory representation of the
            given dataset.
        """
        timeseries = tokio.timeseries.TimeSeries()
        timeseries.dataset_name = dataset_name

        try:
            dataset = self[dataset_name]
        except KeyError:
            # can't attach because dataset doesn't exist; pass this back to caller so it can init
            return None

        timeseries.dataset = dataset if light else dataset[:, :]

        # load and decode version of dataset and file schema
        timeseries.global_version = self['/'].attrs.get('version')
        timeseries.version = self.get_version(dataset_name)
        if isinstance(timeseries.version, bytes):
            timeseries.version = timeseries.version.decode()

        # copy columns into memory
        columns = self.get_columns(dataset_name)
        timeseries.set_columns(columns)

        # copy metadata into memory
        for key, value in dataset.attrs.items():
            if isinstance(value, bytes):
                timeseries.dataset_metadata[key] = value.decode()
            else:
                timeseries.dataset_metadata[key] = value
        for key, value in dataset.parent.attrs.items():
            if isinstance(value, bytes):
                timeseries.group_metadata[key] = value.decode()
            else:
                timeseries.group_metadata[key] = value

        timeseries.timestamp_key = get_timestamps_key(self, dataset_name)
        timeseries.timestamps = self[timeseries.timestamp_key]
        timeseries.timestamps = timeseries.timestamps if light else timeseries.timestamps[:]

        timeseries.timestep = timeseries.timestamps[1] - timeseries.timestamps[0]
        return timeseries

    def commit_timeseries(self, timeseries, **kwargs):
        """Writes contents of a TimeSeries object into a group

        Args:
            timeseries (tokio.timeseries.TimeSeries): the time series to save
                as a dataset within self
            kwargs (dict): Extra arguments to pass to self.create_dataset()
        """
        extra_dataset_args = {
            'dtype': 'f8',
            'chunks': True,
            'compression': 'gzip',
        }
        extra_dataset_args.update(kwargs)

        # Create the dataset in the HDF5 file (if necessary)
        if timeseries.dataset_name in self:
            dataset_hdf5 = self[timeseries.dataset_name]
        else:
            dataset_hdf5 = self.create_dataset(name=timeseries.dataset_name,
                                               shape=timeseries.dataset.shape,
                                               **extra_dataset_args)

        # when timestamp_key has been left empty, use the default
        timestamp_key = timeseries.timestamp_key
        if timestamp_key is None:
            timestamp_key = '/'.join(timeseries.dataset_name.split('/')[0:-1] \
                                 + [DEFAULT_TIMESTAMP_DATASET])

        # Create the timestamps in the HDF5 file (if necessary) and calculate
        # where to insert our data into the HDF5's dataset
        if timestamp_key not in self:
            timestamps_hdf5 = self.create_dataset(name=timestamp_key,
                                                  shape=timeseries.timestamps.shape,
                                                  dtype='i8')
            # Copy the in-memory timestamp dataset into the HDF5 file
            timestamps_hdf5[:] = timeseries.timestamps[:]
            t_start = 0
            t_end = timeseries.timestamps.shape[0]
            start_timestamp = timeseries.timestamps[0]
            end_timestamp = timeseries.timestamps[-1] + timeseries.timestep
        else:
            existing_timestamps = self.get_timestamps(timeseries.dataset_name)
            t_start, t_end = get_insert_indices(timeseries.timestamps, existing_timestamps)

            if t_start < 0 \
            or t_start > (len(existing_timestamps) - 2) \
            or t_end < 1 \
            or t_end > len(existing_timestamps):
                raise IndexError("cannot commit dataset that is not a subset of existing data")

            start_timestamp = existing_timestamps[0]
            end_timestamp = existing_timestamps[-1] + timeseries.timestep

        # Make sure that the start/end timestamps are consistent with the HDF5
        # file's global time range
        if 'start' not in self.attrs:
            self.attrs['start'] = start_timestamp
            self.attrs['end'] = end_timestamp
        else:
            if self.attrs['start'] != start_timestamp \
            or self.attrs['end'] != end_timestamp:
#               warnings.warn(
                raise IndexError("Mismatched start or end values:  %d != %d or %d != %d" % (
                    start_timestamp, self.attrs['start'],
                    end_timestamp, self.attrs['end']))

        # If we're updating an existing dataset, use its column names and ordering.
        # Otherwise sort the columns before committing them.
        if COLUMN_NAME_KEY in dataset_hdf5.attrs:
            timeseries.rearrange_columns(timeseries.columns)
        else:
            timeseries.sort_columns()

        # Copy the in-memory dataset into the HDF5 file
        dataset_hdf5[t_start:t_end, :] = timeseries.dataset[:, :]

        # Copy column names into metadata before committing metadata
        timeseries.dataset_metadata[COLUMN_NAME_KEY] = timeseries.columns
        timeseries.dataset_metadata['updated'] = int(time.mktime(datetime.datetime.now().timetuple()))

        # If timeseries.version was never set, don't set a dataset-level version in the HDF5
        if timeseries.version is not None:
            self.set_version(timeseries.version, dataset_name=timeseries.dataset_name)

        # Set the file's global version to indicate its schema
        if timeseries.global_version is not None:
            self['/'].attrs['version'] = timeseries.global_version

        # Insert/update dataset metadata
        for key, value in timeseries.dataset_metadata.items():
            # special hack for column names
            if key == COLUMN_NAME_KEY:
                # note: the behavior of numpy.string_(x) where
                # type(x) == numpy.array is _different_ in python2 vs. python3.
                # Python3 happily converts each element to a numpy.string_,
                # while Python2 first calls a.__repr__ to turn it into a single
                # string, then converts that to numpy.string_.
                dataset_hdf5.attrs[key] = numpy.array([numpy.string_(x) for x in value])
            elif tokio.common.isstr(value):
                dataset_hdf5.attrs[key] = numpy.string_(value)
            elif value is None:
                warnings.warn("Skipping attribute %s (null value) for %s" % (key, timeseries.dataset_name))
            else:
                dataset_hdf5.attrs[key] = value

        # Insert/update group metadata
        for key, value in timeseries.group_metadata.items():
            if tokio.common.isstr(value):
                dataset_hdf5.parent.attrs[key] = numpy.string_(value)
            else:
                dataset_hdf5.parent.attrs[key] = value

def missing_values(dataset, inverse=False):
    """Identify matrix values that are missing

    Because we initialize datasets with -0.0, we can scan the sign bit of every
    element of an array to determine how many data were never populated.  This
    converts negative zeros to ones and all other data into zeros then count up
    the number of missing elements in the array.

    Args:
        dataset: dataset to access
        inverse (bool): return 0 for missing and 1 for present if True

    Returns:
        numpy.ndarray: Array of numpy.int8 of 1 and 0 to indicate the presence
        or absence of specific elements
    """
    zero = numpy.int8(0)
    one = numpy.int8(1)
    if inverse:
        converter = numpy.vectorize(lambda x:
                                    zero if (x == 0.0 and math.copysign(1, x) < 0.0) else one)
    else:
        converter = numpy.vectorize(lambda x:
                                    one if (x == 0.0 and math.copysign(1, x) < 0.0) else zero)
    return converter(dataset)


def reduce_dataset_name(key):
    """Divide a dataset name into is base and modifier
    Args:
        dataset_name (str): Key to reference a dataset that may or may not have
            a modifier suffix
    Returns:
        tuple of (str, str or None): First string is the base key, the second
            string is the modifier.
    """
    if key.endswith('/missing'):
        return tuple(key.rsplit('/', 1))
    return key, None

def get_insert_indices(my_timestamps, existing_timestamps):
    """
    Given new timestamps and an existing series of timestamps, find the indices
    overlap so that new data can be inserted into the middle of an existing
    dataset
    """
    existing_timestep = existing_timestamps[1] - existing_timestamps[0]
    my_timestep = my_timestamps[1] - my_timestamps[0]

    # make sure the time delta is ok
    if existing_timestep != my_timestep:
        raise Exception("Existing dataset has different timestep (mine=%d, existing=%d)"
                        % (my_timestep, existing_timestep))

    my_offset = (my_timestamps[0] - existing_timestamps[0]) // existing_timestep
    my_end = my_offset + len(my_timestamps)

    return my_offset, my_end



#!/usr/bin/env pthon
"""Common routines used to apply site-specific info to connectors
"""

import os
import datetime

def enumerate_dated_files(start, end, template,
                          lookup_key=None, match_first=True,
                          timedelta=datetime.timedelta(days=1)):
    """Locate existing time-indexed files between a start and end time.

    Given a start time, end time, and template data structure that describes a
    pattern by which the files of interest are indexed, locate all existing
    files that fall between the start and end time.

    The template argument (`template`) are paths that are passed through
    `datetime.strftime` and then checked for existence for every `timedelta`
    increment between `start` and `end`, inclusive.  `template` may be one
    of three data structures:

        - str: search for files matching this template
        - list of str: search for files matching each template.  If
          ``match_first`` is True, only the first hit per list item per time
          interval is returned; otherwise, every file matching every template
          in the entire list is returned.
        - dict: use ``lookup_key`` to determine the element in the dictionary
          to use as the template.  That value is treated as a new ``template``
          object and can be of any of these three types.

    Args:
        start (datetime.datetime): Begin including files corresponding to this
            start date, inclusive.
        end (datetime.datetime): Stop including files with timestamps that
            follow this end date.  Resulting files _will_ include this date.
        template (str, list, or dict): Template string(s) which should be passed
            to datetime.strftime to be converted into specific time-delimited
            files.
        lookup_key (str or None): When `type(template)` is dict, use this key to
            identify the key-value to use as template.  If None and `template` is
            a dict, iterate through all values of `template`.
        match_first (bool): If True, only return the first matching file for
            each time increment checked.  Otherwise, return _all_ matching
            files.
        timedelta (datetime.timedelta): Increment to use when iterating between
            `start` and `end` while looking for matching files.
    """

    if end < start:
        raise IndexError("datetime_end < datetime_start")

    check_paths = _expand_check_paths(template, lookup_key)

    day = start
    results = []
    while day.date() <= end.date():
        results += _match_files(check_paths, day, match_first)
        day += timedelta

    return results


def _expand_check_paths(template, lookup_key):
    """Generate paths to examine from a variable-type template.

    `template` may be one of three data structures:

        - str: search for files matching this exact template
        - list of str: search for files matching each template listed.
        - dict: use `lookup_key` to determine the element in the dictionary
          to use as the template.  That value is treated as a new `template`
          object and can be of any of these three types.

    Args:
        template (str, list, or dict): Template string(s) which should be passed
            to datetime.datetime.strftime to be converted into specific
            time-delimited files.
        lookup_key (str or None): When `type(template)` is dict, use this key
            to identify the key-value to use as template.  If None and
            ``template`` is a dict, iterate through all values of `template`.
    """
    check_paths = []
    if isinstance(template, dict):
        if lookup_key is None:
            check_paths += _expand_check_paths(list(template.values()), lookup_key)
        else:
            check_paths += _expand_check_paths(template.get(lookup_key, []), lookup_key)
    elif isinstance(template, list):
        for value in template:
            check_paths += _expand_check_paths(value, lookup_key)
    else:
        check_paths += [template]

    return check_paths


def _match_files(check_paths, use_time, match_first):
    """Locate file(s) that match a templated file path for a given time

    Args:
        check_paths (list of str): List of templates to pass to strftime
        use_time (datetime.datetime): Time to pass through strftime to generate
            an actual file path to check for existence.
        match_first (bool): If True, only return the first matching file for
            each time increment checked.  Otherwise, return _all_ matching
            files.
    """

    matching = []
    for check_path in check_paths:
        match_path = use_time.strftime(check_path)
        if os.path.exists(match_path):
            matching.append(match_path)
            if match_first:
                break

    return matching



#!/usr/bin/env python
"""Retrieve data generated by collectd and stored in Elasticsearch
"""

import copy
from . import es

BASE_QUERY = copy.deepcopy(es.BASE_QUERY)
es.mutate_query(BASE_QUERY, term='prefix', field='hostname', value='bb')

QUERY_DISK_DATA = copy.deepcopy(BASE_QUERY)
es.mutate_query(QUERY_DISK_DATA, term='term', field='plugin', value='disk')
es.mutate_query(QUERY_DISK_DATA, term='prefix', field='plugin_instance', value='nvme')
es.mutate_query(QUERY_DISK_DATA, term='regexp', field='collectd_type', value='disk_(octets|ops)')

QUERY_CPU_DATA = copy.deepcopy(BASE_QUERY)
es.mutate_query(QUERY_CPU_DATA, term='term', field='plugin', value='cpu')
es.mutate_query(QUERY_CPU_DATA, term='regexp', field='type_instance', value='(idle|user|system)')

QUERY_MEMORY_DATA = copy.deepcopy(BASE_QUERY)
es.mutate_query(QUERY_MEMORY_DATA, term='term', field='plugin', value='memory')

### Only return the following _source fields
SOURCE_FILTER = [
    '@timestamp',
    'hostname',
    'plugin',
    'collectd_type',
    'type_instance',
    'plugin_instance',
    'value',
    'longterm',
    'midterm',
    'shortterm',
    'majflt',
    'minflt',
    'if_octets',
    'if_packets',
    'if_errors',
    'rx',
    'tx',
    'read',
    'write',
    'io_time',
]


class CollectdEs(es.EsConnection):
    """collectd-Elasticsearch connection handler
    """
    def __init__(self, *args, **kwargs):
        super(CollectdEs, self).__init__(*args, **kwargs)
        self.filter_function = lambda x: x['hits']['hits']
        self.flush_every = 50000
        self.flush_function = lambda x: x

    @classmethod
    def from_cache(cls, *args, **kwargs):
        instance = super(CollectdEs, cls).from_cache(*args, **kwargs)
        instance.filter_function = lambda x: x['hits']['hits']
        instance.flush_every = 50000
        instance.flush_function = lambda x: x
        return instance

    def query_disk(self, start, end):
        """Query Elasticsearch for collectd disk plugin data.

        Args:
            start (datetime.datetime): lower bound for query (inclusive)
            end (datetime.datetime): upper bound for query (exclusive)
        """
        self.query_timeseries(QUERY_DISK_DATA, start, end)

    def query_memory(self, start, end):
        """Query Elasticsearch for collectd memory plugin data.

        Args:
            start (datetime.datetime): lower bound for query (inclusive)
            end (datetime.datetime): upper bound for query (exclusive)
        """
        self.query_timeseries(QUERY_MEMORY_DATA, start, end)

    def query_cpu(self, start, end):
        """Query Elasticsearch for collectd cpu plugin data.

        Args:
            start (datetime.datetime): lower bound for query (inclusive)
            end (datetime.datetime): upper bound for query (exclusive)
        """
        self.query_timeseries(QUERY_CPU_DATA, start, end)

    def query_timeseries(self, query_template, start, end, source_filter=None,
                         filter_function=None, flush_every=None,
                         flush_function=None):
        """Map connection-wide attributes to super(self).query_timeseries arguments

        Args:
            query_template (dict): a query object containing at least one
                ``@timestamp`` field
            start (datetime.datetime): lower bound for query (inclusive)
            end (datetime.datetime): upper bound for query (exclusive)
            source_filter (bool or list): Return all fields contained in each
                document's _source field if True; otherwise, only return source
                fields contained in the provided list of str.  If None, use the
                default for this connector.
            filter_function (function, optional): Function to call before each
                set of results is appended to the ``scroll_pages`` attribute; if
                specified, return value of this function is what is appended.
                If None, use the default for this connector.
            flush_every (int or None): trigger the flush function once the
                number of docs contained across all ``scroll_pages`` reaches
                this value.  If None, do not apply `flush_function`.  If None,
                use the default for this connector.
            flush_function (function, optional): function to call when
                `flush_every` docs are retrieved.  If None, use the default for
                this connector.
        """
        if source_filter is None:
            source_filter = SOURCE_FILTER
        if filter_function is None:
            filter_function = self.filter_function
        if flush_every is None:
            flush_every = self.flush_every
        if flush_function is None:
            flush_function = self.flush_function
        return super(CollectdEs, self)\
            .query_timeseries(query_template=query_template,
                              start=start,
                              end=end,
                              source_filter=source_filter,
                              filter_function=filter_function,
                              flush_every=flush_every,
                              flush_function=flush_function)

    def to_dataframe(self):
        """Converts self.scroll_pages to a DataFrame

        Returns:
            pandas.DataFrame: Contents of the last query's pages
        """
        return super(CollectdEs, self).to_dataframe(fields=SOURCE_FILTER)



"""Provides interfaces into ESnet's SNMP REST API

Documentation for the REST API is here:

    http://es.net/network-r-and-d/data-for-researchers/snmp-api/

Notes:
    This connector relies either on the ``esnet_snmp_uri`` configuration value
    being set in the pytokio configuration or the ``PYTOKIO_ESNET_SNMP_URI``
    being defined in the runtime environment.

Examples:
    Retrieving the data of multiple endpoints (ESnet routers) and interfaces
    is a common pattern.  To do this, the ``EsnetSnmp`` object should be
    initialized with only the intended start/end times, and the object should
    be asynchronously populated using calls to
    ``EsnetSnmp.get_interface_counters``::

        import datetime
        import tokio.connectors.esnet_snmp

        ROUTER = 'sunn-cr5'
        INTERFACE = 'to_nersc_ip-d_v4'
        TARGET_DATE = datetime.datetime.today() - datetime.timedelta(days=1)

        # Because the ESnet API treats the end date as inclusive, we subtract
        # one second to avoid counting the first measurement of the following
        # day.
        esnetsnmp = tokio.connectors.esnet_snmp.EsnetSnmp(
            start=TARGET_DATE,
            end=TARGET_DATE + datetime.timedelta(days=1, seconds=-1))
        for direction in 'in', 'out':
            esnetsnmp.get_interface_counters(
                endpoint=ROUTER,
                interface=INTERFACE,
                direction=direction,
                agg_func='average')

        for direction in 'in', 'out':
            bytes_per_sec = list(esnetsnmp[ROUTER][INTERFACE][direction].values())
            total_bytes = sum(bytes_per_sec) * esnetsnmp.timestep
            print("%s:%s saw %.2f TiB %s" % (
                ROUTER,
                INTERFACE,
                total_bytes / 2**40,
                direction))

    For simple queries, it is sufficient to specify the endpoint, interface,
    and direction directly in the initialization::

        esnetsnmp = tokio.connectors.esnet_snmp.EsnetSnmp(
            start=TARGET_DATE,
            end=TARGET_DATE + datetime.timedelta(days=1, seconds=-1)
            endpoint=ROUTER,
            interface=INTERFACE,
            direction="in")
        print("Total bytes in: %.2f" % (
            sum(list(esnetsnmp[ROUTER][INTERFACE]["in"].values())) / 2**40))

"""

import time
import json
import datetime
import warnings

import requests
import pandas

from .. import config
from . import common

class EsnetSnmp(common.CacheableDict):
    """Container for ESnet SNMP counters

    Dictionary with structure::

        {
            "endpoint0": {
                "interface_x": {
                    "in": {
                        timestamp1: value1,
                        timestamp2: value2,
                        timestamp3: value3,
                        ...
                    },
                    "out": { ... }
                },
                "interface_y": { ... }
            },
            "endpoint1": { ... }
        }

    Various methods are provided to access the data of interest.
    """
    def __init__(self,
                 start,
                 end,
                 endpoint=None,
                 interface=None,
                 direction=None,
                 agg_func=None,
                 interval=None,
                 **kwargs):
        """Retrieves data rate data for an ESnet endpoint

        Initializes the object with a start and end time.  Optionally runs a
        REST API query if endpoint, interface, and direction are provided.
        Assumes that once the start/end time have been specified, they should
        not be changed.

        Args:
            start (datetime.datetime): Start of interval to retrieve, inclusive
            end (datetime.datetime): End of interval to retrieve, inclusive
            endpoint (str, optional): Name of the ESnet endpoint whose data is
                being retrieved
            interface (str, optional): Name of the ESnet endpoint interface on
                the specified endpoint
            direction (str, optional): "in" or "out" to signify data input into
                ESnet or data output from ESnet
            agg_func (str, optional ): Specifies the reduction operator to be
                applied over each interval; must be one of "average," "min," or
                "max."  If None, uses the ESnet default.
            interval (int, optional ): Resolution, in seconds, of the data to be
                returned.  If None, uses the ESnet default.
            kwargs (dict): arguments to pass to super.__init__()

        Attributes:
            start (datetime.datetime): Start of interval represented by this object, inclusive
            end (datetime.datetime): End of interval represented by this object, inclusive
            start_epoch (int): Seconds since epoch for self.start
            end_epoch (int): Seconds since epoch for self.end
        """

        super(EsnetSnmp, self).__init__(**kwargs)

        if start >= end:
            raise RuntimeError("Start must be greater than or equal to end")

        # input parameters
        self.start = start
        self.end = end
        self.start_epoch = int(time.mktime(self.start.timetuple()))
        self.end_epoch = int(time.mktime(self.end.timetuple()))
        self.requested_timestep = interval
        self.requested_endpoint = endpoint
        self.requested_interface = interface
        self.requested_direction = direction
        self.requested_agg_func = agg_func

        # output parameters
        self.last_response = None
        self.timestep = None

        if endpoint and interface and direction:
            self.get_interface_counters(endpoint=endpoint,
                                        interface=interface,
                                        direction=direction,
                                        agg_func=agg_func,
                                        interval=interval)

    def load_json(self, **kwargs):
        """Loads input from serialized JSON

        Need to coerce timestamp keys back into ints from strings
        """
        super(EsnetSnmp, self).load_json(**kwargs)

        for endpoint, interfaces in self.items():
            for interface, directions in interfaces.items():
                for direction, data in directions.items():
                    if isinstance(data, dict):
                        keys = list(data.keys())
                        # make a copy of the key as an int
                        for key in keys:
                            data[int(key)] = data[key]
                        # delete the old non-int key
                        for key in keys:
                            del data[key]

    def _insert_result(self):
        """Parse the raw output of the REST API and update self

        ESnet's REST API will return an object like::

            {
                "agg": "30",
                "begin_time": 1517471100,
                "end_time": 1517471910,
                "cf": "average",
                "data": [
                    [
                        1517471100,
                        5997486471.266666
                    ],
            ...
                    [
                        1517471910,
                        189300026.8
                    ]
                ]
            }

        Args:
            result (dict): JSON structure returned by the ESnet REST API

        Returns:
            bool: True if data inserted without errors; False otherwise
        """

        result = self.last_response
        if not result:
            print("no result; bailing")
            return False

        # set the timestep if not previously set
        interval = _get_interval_result(result)
        if interval:
            if not self.timestep:
                self.timestep = interval
            elif self.timestep != interval:
                warnings.warn("received timestep %d from an object with timestep %d"
                              % (interval, self.timestep))

        if self.requested_endpoint not in self:
            self[self.requested_endpoint] = {}
        if self.requested_interface not in self[self.requested_endpoint]:
            self[self.requested_endpoint][self.requested_interface] = {}
        if self.requested_direction not in self[self.requested_endpoint][self.requested_interface]:
            self[self.requested_endpoint][self.requested_interface][self.requested_direction] = {}

        data = result.get('data')
        if not data:
            warnings.warn("No data in result")

        for timestamp, value in data:
            self[self.requested_endpoint][self.requested_interface][self.requested_direction][timestamp] = value

        self[self.requested_endpoint][self.requested_interface]['units'] = 'bytes/sec'

        return True

    def get_interface_counters(self, endpoint, interface, direction, agg_func=None, interval=None):
        """Retrieves data rate data for an ESnet endpoint

        Args:
            endpoint (str): Name of ESnet endpoint (usually a router identifier)
            interface (str): Name of the ESnet endpoint interface
            direction (str): "in" or "out" to signify data input into ESnet or
                data output from ESnet
            agg_func (str or None): Specifies the reduction operator to be
                applied over each interval; must be one of "average," "min," or
                "max."  If None, uses the ESnet default.
            interval (int or None): Resolution, in seconds, of the data to be
                returned.  If None, uses the ESnet default.

        Returns:
           dict: raw return from the REST API call
        """
        if direction not in ['in', 'out']:
            raise RuntimeError("direction must be either in or out")

        self.requested_endpoint = endpoint
        self.requested_interface = interface
        self.requested_direction = direction
        self.requested_agg_func = agg_func
        self.requested_timestep = interval

        if self.requested_timestep is not None \
        and self.timestep is not None \
        and self.requested_timestep != self.timestep:
            warnings.warn("received timestep %d from an object with timestep %d"
                          % (self.requested_timestep, self.timestep))

        uri = config.CONFIG.get('esnet_snmp_uri')
        if uri is None:
            raise requests.exceptions.ConnectionError("no esnet_snmp_uri configured")
        uri += '/%s/interface/%s/%s' % (endpoint, interface, direction)

        params = {
            'begin': self.start_epoch,
            'end': self.end_epoch,
        }
        if agg_func is not None:
            params['calc_func'] = agg_func
        if interval is not None:
            params['calc'] = interval

        request = requests.get(uri, params=params)
        request.raise_for_status()
        self.last_response = request.json()

        self._insert_result()

        return self.last_response

    def to_dataframe(self, multiindex=False):
        """Return data as a Pandas DataFrame

        Args:
            multiindex (bool): If True, return a DataFrame indexed by timestamp,
                endpoint, interface, and direction
        """

        to_df = []

        for endpoint, interfaces in self.items():
            for interface, directions in interfaces.items():
                for direction, data in directions.items():
                    if isinstance(data, dict):
                        for timestamp, value in data.items():
                            to_df.append({
                                'endpoint': endpoint,
                                'interface': interface,
                                'direction': direction,
                                'timestamp': datetime.datetime.fromtimestamp(timestamp),
                                'data_rate': value,
                            })

        dataframe = pandas.DataFrame.from_records(to_df)
        if multiindex:
            return dataframe.set_index(['timestamp', 'endpoint', 'interface', 'direction'])

        return dataframe

def _get_interval_result(result):
    """Parse the raw output of the REST API output and return the timestep

    Args:
        result (dict): the raw JSON output of the ESnet REST API
    """
    value = result.get('agg')
    if value:
        return int(value)

    value = result.get('calc')
    if 'calc' in result:
        return int(value)

    return None



#!/usr/bin/env python
"""
This module provides generic infrastructure for retrieving data from a
relational database that contains immutable data.  It can use a local caching
database (sqlite3) to allow for reanalysis on platforms that cannot access the
original remote database or to reduce the load on remote databases.
"""

import warnings
try:
    import pymysql
    pymysql.install_as_MySQLdb()
except ImportError:
    pass
try:
    import MySQLdb
except ImportError:
    pass

import sqlite3

HIT_CACHE_DB = 1
HIT_REMOTE_DB = 2

class CachingDb(object):
    """Connect relational database with an optional caching layer interposed.
    """
    #pylint: disable=too-many-arguments
    def __init__(self, dbhost=None, dbuser=None, dbpassword=None, dbname=None, cache_file=None):
        """Connect to a relational database.

        If instantiated with a cache_file argument, all queries will go to that
        SQLite-based cache database.  If this class is not instantiated with a
        cache_file argument, all queries will go out to the remote database.

        If none of the connection arguments (``db*``) are specified, do not
        connect to a remote database and instead rely entirely on the caching
        database or a separate call to the ``connect()`` method.

        Arguments:
            dbhost (str, optional): hostname for the remote database
            dbuser (str, optional): username to use when connecting to database
            dbpassword (str, optional): password for authenticating to database
            dbname (str, optional): name of database to use when connecting
            cache_file (str, optional):  Path to an SQLite3 database to use as
                a caching layer.

        Attributes:
            saved_results (dict): in-memory data cache, keyed by table names
                and whose values are dictionaries with keys ``rows`` and
                ``schema``.  ``rows`` are a list of row tuples returned
                from earlier queries, and ``schema`` is the SQL statement
                required to create the table corresponding to ``rows``.
            last_hit (int): a flag to indicate whether the last query was
                found in the caching database or the remote database
            cache_file (str): path to the caching database's file
            cache_db (sqlite3.Connection): caching database connection handle
            cache_db_ps (str): paramstyle of the caching database as defined
                by `PEP-0249`_
            remote_db: remote database connection handle
            remote_db_ps (str): paramstyle of the remote database as defined
                by `PEP-0249`_

        .. _PEP-0249: https://www.python.org/dev/peps/pep-0249
        """

        # self.saved_results is the in-memory data cache.  It has a structure of
        # saved_results = {
        #    'table1': {
        #       'rows': [ (row1), (row2), (row3), ... ],
        #       'schema': 'create table if not exists table1(...)',
        #    }
        #    'table2': { ... }
        #    ...
        # }
        self.saved_results = {}
        self.last_hit = None

        # cache db
        self.cache_file = None
        self.cache_db = None
        self.cache_db_ps = None

        # actual db
        self.remote_db = None
        self.remote_db_ps = None

        # Connect to cache db if specified
        if cache_file is not None:
            self.connect_cache(cache_file)
            # TODO: check to ensure db isn't empty
#           result = self.query('select * from summary limit 1', (), nocache=True)
#           if len(result) == 0:
#               warnings.warn("Using an empty cache database; queries will return nothing")

        if dbhost is not None \
        and dbuser is not None \
        and dbpassword is not None \
        and dbname is not None:
            self.connect(dbhost=dbhost,
                         dbuser=dbuser,
                         dbpassword=dbpassword,
                         dbname=dbname)

    def connect(self, dbhost, dbuser, dbpassword, dbname):
        """Establish remote db connection.

        Connects to a remote MySQL database and defines the connection handler
        and paramstyle attributes.

        Args:
            dbhost (str): hostname for the remote database
            dbuser (str): username to use when connecting to database
            dbpassword (str): password for authenticating to database
            dbname (str): name of database to use when connecting
        """
        if self.cache_db is not None:
            # can't really do both local and remote dbs; all queries will be
            # run against the cache_db, which is probably not what someone
            # wants
            warnings.warn("attempting to use both remote and cache db; disabling cache db")
            self.close_cache()

        self.remote_db = MySQLdb.connect(host=dbhost,
                                         user=dbuser,
                                         passwd=dbpassword,
                                         db=dbname)
        self.remote_db_ps = get_paramstyle_symbol(MySQLdb.paramstyle)

    def close(self):
        """Destroy connection objects.

        Close the remote database connection handler and reset state of remote
        connection attributes.
        """
        self.remote_db = None
        self.remote_db_ps = None

    def connect_cache(self, cache_file):
        """Open the cache database file and set the handler attribute.

        Args:
            cache_file (str): Path to the SQLite3 caching database file to be
                used.
        """
        if cache_file is not None:
            self.cache_db = sqlite3.connect(cache_file)
            self.cache_file = cache_file
            self.cache_db_ps = get_paramstyle_symbol(sqlite3.paramstyle)

    def close_cache(self):
        """Close the cache database handler and reset caching db attributes.
        """
        if self.cache_db is not None:
            self.cache_db = self.cache_db.close()
            self.cache_file = None

    def drop_cache(self, tables=None):
        """Flush saved results from memory.

        If tables are specified, only drop those tables' results.  If no tables
        are provided, flush everything.

        Args:
            tables (list, optional): List of table names (str) to flush.  If
                omitted, flush all tables in cache.
        """
        drop_caches = set([])
        for table in self.saved_results:
            if tables is None or table in tables:
                drop_caches.add(table)

        for drop_cache in drop_caches:
            del self.saved_results[drop_cache]

    def save_cache(self, cache_file):
        """Commit the in-memory cache to a cache database.

        This method is currently very memory-inefficient and not good for
        caching giant pieces of a database without something wrapping it to feed
        it smaller pieces.

        Note:
            This manipulates the ``cache_db*`` attributes in a dirty way
            to prevent closing and re-opening the original cache db.  If the
            ``self.open_cache()`` is ever changed to include tracking more
            state, this function must also be updated to retain that state while
            the old cache db state is being temporarily shuffled out.

        Args:
            cache_file (str): Path to the cache file to be used to write out
                the cache contents.  This file will temporarily pre-empt the
                `cache_file` attribute and should be a different file.
        """
        ### Shuffle out the old cache db state (if it exists)
        old_state = {}
        if self.cache_file is not None and self.cache_file != cache_file:
            old_state = {
                'cache_file': self.cache_file,
                'cache_db': self.cache_db,
                'cache_db_ps': self.cache_db_ps,
            }

        ### Open a new cache db connection without closing the old cache db
        self.connect_cache(cache_file)

        ### Commit each table we've retained in memory
        drop_caches = set([])
        for table, table_info in self.saved_results.items():
            if len(self.saved_results[table]['rows']) < 1:
                warnings.warn("table %s has no rows" % table)
                continue

            num_fields = None
            ### Verify and preprocess each saved row
            for row in self.saved_results[table]['rows']:
                if num_fields is None:
                    num_fields = len(row)

                ### Verify that the rows we've saved are actually all of the
                ### same length so that they have a hope of being inserted
                ### into the schema
                if len(row) != num_fields:
                    warnings.warn(
                        "saved_results[%s] contains non-uniform rows (%d, %d); skipping table" %
                        table, len(row), num_fields)
                    continue
                else:
                    ### Prepend table name to row to facilitate the bulk insert
                    ### query below
#                   self.saved_results[table]['rows'][index] = (table,) + row
                    pass

            ### Create the table (if necessary).  This will throw all sorts of
            ### exceptions if
            ###   (1) the table doesn't already exist in the cache database, or
            ###   (2) 'schema' isn't set correctly by the downstream application
            if table_info['schema'] is not None:
                self.cache_db.execute(
                    "CREATE TABLE IF NOT EXISTS %s (%s, PRIMARY KEY(%s))" %
                    (table,
                     ', '.join(table_info['schema']['columns']),
                     ', '.join(table_info['schema']['primary_key'])))

            ### INSERT OR REPLACE so that the cache db never wins if a duplicate
            ### primary key is detected
            query_str = "insert or replace into %s values (%s)" % (table, ','.join(['?'] * num_fields))
            self.cache_db.executemany(
                query_str,
                table_info['rows'])
            self.cache_db.commit()

            ### Drop committed rows from memory
            drop_caches.add(table)

        for drop_cache in drop_caches:
            del self.saved_results[drop_cache]

        self.close_cache()

        ### Shuffle back in the state of the old cache db
        if len(old_state) > 0:
            self.cache_file = old_state['cache_file']
            self.cache_db = old_state['cache_db']
            self.cache_db_ps = old_state['cache_db_ps']

    def query(self, query_str, query_variables=(), table=None, table_schema=None):
        """Pass a query through all layers of cache and return on the first hit.

        If a table is specified, the results of this query can be saved to the
        cache db into a table of that name.

        Args:
            query_str (str): SQL query expressed as a string
            query_variables (tuple): parameters to be substituted into
                `query_str` if `query_str` is a parameterized query
            table (str, optional): name of table in the cache database to save
                the results of the query
            table_schema (str, optional): when `table` is specified, the SQL
                line to initialize the table in which the query results will
                be cached.

        Returns:
            tuple: Tuple of tuples corresponding to rows of fields as returned
            by the SQL query.
        """

        ### Collapse query string to remove extraneous whitespace
        query_str = ' '.join(query_str.split())

        ### Check the cache database (if available)
        if self.cache_db is not None:
            results = self._query_sqlite3(query_str, query_variables)
            self.last_hit = HIT_CACHE_DB
        ### Check the MySQL database (if available)
        elif self.remote_db is not None:
            results = self._query_mysql(query_str, query_variables)
            self.last_hit = HIT_REMOTE_DB
        else:
            raise RuntimeError('No databases available to query')

        if table is not None:
            ### Initialize the table if our intent is to save the result of this
            ### query.
            if table not in self.saved_results:
                self.saved_results[table] = {
                    'rows': [],
                    'schema': None,
                }
            ### Table schema can be defined or re-defined on any query.  It is
            ### up to the downstream application to manage this correctly.
            if table_schema is not None:
                self.saved_results[table]['schema'] = table_schema

            ### Append our results
            self.saved_results[table]['rows'] += list(results)

        return results

    def _query_sqlite3(self, query_str, query_variables):
        """Run a query against the cache database and return the full output.

        Args:
            query_str (str): SQL query expressed as a string
            query_variables (tuple): parameters to be substituted into
                `query_str` if `query_str` is a parameterized query
        """
        cursor = self.cache_db.cursor()
        if '%(ps)' in query_str:
            query_str = query_str % {'ps': self.cache_db_ps}
        cursor.execute(query_str, query_variables)
        rows = cursor.fetchall()
        cursor.close()
        return rows

    def _query_mysql(self, query_str, query_variables):
        """Run a query against the MySQL database and return the full output.

        Args:
            query_str (str): SQL query expressed as a string
            query_variables (tuple): parameters to be substituted into
                `query_str` if `query_str` is a parameterized query
        """
        cursor = self.remote_db.cursor()
        if '%(ps)' in query_str:
            query_str = query_str % {'ps': self.remote_db_ps}
        cursor.execute(query_str, query_variables)
        rows = cursor.fetchall()
        cursor.close()
        return rows

def get_paramstyle_symbol(paramstyle):
    """Infer the correct paramstyle for a database.paramstyle

    Provides a generic way to determine the paramstyle of a database connection
    handle.  See `PEP-0249`_ for more information.

    Args:
        paramstyle (str): Result of a generic database handler's `paramstyle`
            attribute

    Returns:
        str: The string corresponding to the paramstyle of the given database
        connection.

    .. _PEP-0249: https://www.python.org/dev/peps/pep-0249
    """
    if paramstyle == 'qmark':
        return "?"
    elif paramstyle == 'format' or paramstyle == 'pyformat':
        return "%s"
    else:
        raise Exception("Unsupported paramstyle %s" % paramstyle)



"""Helper classes and functions used by the HDF5 connector

This contains some of the black magic required to make older H5LMT files
compatible with the TOKIO HDF5 schemas and API.
"""

import numpy
import h5py

TIMESTAMP_KEY = 'timestamps'
DEFAULT_TIMESTAMP_DATASET = 'timestamps' # this CANNOT be an absolute location
COLUMN_NAME_KEY = 'columns'

class MappedDataset(h5py.Dataset):
    """
    h5py.Dataset that applies a function to the results of __getitem__
    before returning the data.  Intended to dynamically generate certain
    datasets that are simple derivatives of others.
    """
    def __init__(self, map_function=None, map_kwargs=None, transpose=False, force2d=False,
                 *args, **kwargs):
        """Configure a MappedDatset

        Attach a map function to a h5py.Dataset (or derivative) and store the
        arguments to be fed into that map function whenever this object gets
        sliced.

        Args:
            map_function (function): function to be called on the value returned
                when parent class is sliced
            map_kwargs (dict): kwargs to be passed into map_function
            transpose (bool): when True, transpose the results of map_function
                before returning them.  Required by some H5LMT datasets.
            force2d (bool): when True, convert a 1d array into a 2d array with
                a single column.  Required by some H5LMT datasets.
        """
        if map_kwargs is None:
            map_kwargs = {}

        super(MappedDataset, self).__init__(*args, **kwargs)

        self.map_function = map_function
        self.map_kwargs = map_kwargs
        self.transpose = transpose
        self.force2d = force2d

    def __getitem__(self, key):
        """
        Apply the map function to the result of the parent class and return that
        transformed result instead.  Transpose is very ugly, but required for
        h5lmt support.
        """

        # The following transformations require the entire dataset to be
        # retrieved before it can be sliced, so retrieve it into a memory
        # buffer
        if self.transpose or self.force2d:
            array_buf = numpy.zeros(shape=self.shape, dtype=self.dtype)
            self.read_direct(array_buf)
            if self.transpose:
                array_buf = array_buf.T
            if self.force2d and len(array_buf.shape) == 1:
                array_buf = array_buf.reshape((array_buf.shape[0], 1))
            # We have to __getitem__ *after* applying the transformation or else
            # we won't get transformed indices
            if self.map_function:
                return self.map_function(array_buf, **self.map_kwargs).__getitem__(key)
            else:
                return array_buf.__getitem__(key)
        else:
            # if we didn't have to preload the whole dataset, we get __getitem__
            # then apply the map function
            result = super(MappedDataset, self).__getitem__(key)
            if self.map_function:
                return self.map_function(result, **self.map_kwargs)
            else:
                return result

def _apply_timestep(return_value, parent_dataset, func=lambda x, timestep: x * timestep):
    """Apply a transformation function to a return value

    Transforms the data returned when slicing a h5py.Dataset object by
    applying a function to the dataset's values.  For example if return_value
    are 'counts per timestep' and you want to convert to 'counts per second',
    you would specify func=lambda x, y: x * y

    Args:
        return_value: the value returned when slicing h5py.Dataset
        parent_dataset: the h5py.Dataset which generated return_value
        func: a function which takes two arguments: the first is return_value,
            and the second is the timestep of parent_dataset

    Returns:
        A modified version of return_value (usually a numpy.ndarray)
    """
    hdf5_file = parent_dataset.file
    dataset_name = parent_dataset.name
    timestamps = get_timestamps(hdf5_file, dataset_name)

    if timestamps is None:
        errmsg = "Could not find timestamps for %s in %s" % (dataset_name, hdf5_file.filename)
        raise KeyError(errmsg)

    timestep = timestamps[1] - timestamps[0]

    return func(return_value, timestep)

def _one_column(return_value, col_idx, apply_timestep_func=None, parent_dataset=None):
    """Extract a specific column from a dataset

    Args:
        return_value: the value returned by the parent DataSet object that we
            will modify
        col_idx: the column index for the column we are demultiplexing
        apply_timestep_func (function): if provided, apply this function with
            return_value as the first argument and the timestep of
            parent_dataset as the second.
        parent_dataset (Dataset): if provided, indicates that return_value
            should be divided by the timestep of parent_dataset to convert
            values to rates before returning

    Returns:
        A modified version of return_value (usually a numpy.ndarray)
    """
    modified_values = return_value[:, col_idx:col_idx+1]
    if parent_dataset and apply_timestep_func:
        modified_values = _apply_timestep(modified_values, parent_dataset, func=apply_timestep_func)
    return modified_values

def convert_counts_rates(hdf5_file, from_key, to_rates, *args, **kwargs):
    """Convert a dataset between counts/sec and counts/timestep

    Retrieve a dataset from an HDF5 file, convert it to a MappedDataset, and
    attach a multiply/divide function to it so that subsequent slices return
    a transformed set of data.

    Args:
        hdf5_file (h5py.File): object from which dataset should be loaded
        from_key (str): dataset name key to load from hdf5_file
        to_rates (bool): convert from per-timestep to per-sec (True) or per-sec
            to per-timestep (False)

    Returns:
        A MappedDataset configured to convert to/from rates when dereferenced
    """
    if from_key not in hdf5_file:
        errmsg = "Could not find dataset_name %s in %s" % (from_key, hdf5_file.filename)
        raise KeyError(errmsg)

    dataset = hdf5_file[from_key]
    map_kwargs = {'parent_dataset': dataset}
    if to_rates:
        map_kwargs['func'] = lambda x, timestep: x / timestep
    else:
        map_kwargs['func'] = lambda x, timestep: x * timestep

    return MappedDataset(bind=dataset.id,
                         map_function=_apply_timestep,
                         map_kwargs=map_kwargs,
                         *args,
                         **kwargs)

def map_dataset(hdf5_file, from_key, *args, **kwargs):
    """Create a MappedDataset

    Creates a MappedDataset from an h5py.File (or derivative).  Functionally
    similar to h5py.File.__getitem__.

    Args:
        hdf5_file (h5py.File or connectors.hdf5.Hdf5): file containing dataset of interest
        from_key (str): name of dataset to apply mapping function to
    """
    if from_key not in hdf5_file:
        errmsg = "Could not find dataset_name %s in %s" % (from_key, hdf5_file.filename)
        raise KeyError(errmsg)

    return MappedDataset(bind=hdf5_file[from_key].id,
                         map_function=None,
                         map_kwargs={},
                         *args,
                         **kwargs)

def demux_column(hdf5_file, from_key, column, apply_timestep_func=None, *args, **kwargs):
    """Extract a single column from an HDF5 dataset

    MappedDataset map function to present a single column from a dataset as an
    entire dataset.  Required to bridge the h5lmt metadata table (which encodes
    all metadata ops in a single dataset) and the TOKIO HDF5 format (which
    encodes a single metadata op per dataset)

    Args:
        hdf5_file (h5py.File): the HDF5 file containing the dataset of interest
        from_key (str): the dataset name from which a column should be extracted
        column (str): the column heading to be returned
        transpose (bool): transpose the dataset before returning it

    Returns:
        A MappedDataset configured to extract a single column when dereferenced
    """
    if from_key not in hdf5_file:
        errmsg = "Could not find dataset_name %s in %s" % (from_key, hdf5_file.filename)
        raise KeyError(errmsg)

    column_idx = list(hdf5_file.get_columns(from_key.lstrip('/'))).index(column)
    map_kwargs = {'col_idx': column_idx}
    if apply_timestep_func:
        map_kwargs['parent_dataset'] = hdf5_file[from_key]
        map_kwargs['apply_timestep_func'] = apply_timestep_func

    return MappedDataset(bind=hdf5_file[from_key].id,
                         map_function=_one_column,
                         map_kwargs=map_kwargs,
                         *args,
                         **kwargs)

def get_timestamps_key(hdf5_file, dataset_name):
    """
    Read into an HDF5 file and extract the name of the dataset containing the
    timestamps correspond to the given dataset_name
    """
    # Look for special 'missing' dataset hack
    if len(dataset_name.strip('/').split('/')) == 3:
        dataset_name = dataset_name.rsplit('/', 1)[0]

    # Get dataset out of HDF5 file.  If dataset doesn't exist, throw exception
    hdf5_dataset = hdf5_file[dataset_name]

    if hdf5_file.attrs.get('version') is None and '/FSStepsGroup/FSStepsDataSet' in hdf5_file:
        return '/FSStepsGroup/FSStepsDataSet'

    # Identify the dataset containing timestamps for this dataset
    if TIMESTAMP_KEY in hdf5_dataset.attrs:
        timestamp_key = hdf5_dataset.attrs[TIMESTAMP_KEY]
    else:
        timestamp_key = hdf5_dataset.parent.name + '/' + DEFAULT_TIMESTAMP_DATASET

    # Load timestamps dataset into memory
    if timestamp_key not in hdf5_file:
        raise KeyError("timestamp_key %s does not exist" % timestamp_key)

    return timestamp_key

def get_timestamps(hdf5_file, dataset_name):
    """
    Return the timestamps dataset for a given dataset name
    """
    return hdf5_file[get_timestamps_key(hdf5_file, dataset_name)]



#!/usr/bin/env python
"""Retrieve Globus transfer logs from NERSC's Elasticsearch infrastructure

Connects to NERSC's OMNI service and retrieves Globus transfer logs.
"""

import copy
from . import es

QUERY = {
    "query": {
        "constant_score": {
            "filter": {
                "bool": {
                    "must": [
                        {
                            "range": {
                                "start_date": {}
                            }
                        },
                        {
                            "range": {
                                "end_date": {}
                            }
                        }
                    ]
                }
            }
        }
    }
}

### Only return the following _source fields
SOURCE_FILTER = [
    '@timestamp',
    'BLOCK',
    'BUFFER',
    'CODE',
    'DATE',
    'DEST',
    'DESTIP',
    'FILE',
    'HOST',
    'NBYTES',
    'START',
    'STREAMS',
    'STRIPES',
    'TASKID',
    'TYPE',
    'USER',
    'VOLUME',
    'bandwidth_mbps',
    'duration',
    'start_date',
    'end_date',
    'host',
]

class NerscGlobusLogs(es.EsConnection):
    """Connection handler for NERSC Globus transfer logs
    """
    def __init__(self, *args, **kwargs):
        super(NerscGlobusLogs, self).__init__(*args, **kwargs)
        self.filter_function = lambda x: x['hits']['hits']
        self.flush_every = 50000
        self.flush_function = lambda x: x

    @classmethod
    def from_cache(cls, *args, **kwargs):
        instance = super(NerscGlobusLogs, cls).from_cache(*args, **kwargs)
        instance.filter_function = lambda x: x['hits']['hits']
        instance.flush_every = 50000
        instance.flush_function = lambda x: x
        return instance

    def query(self, start, end, must=None, scroll=True):
        """Queries Elasticsearch for Globus logs

        Accepts a start time, end time, and an optional "must" field which can
        be used to apply additional term queries.  For example, ``must`` may be::

            [
                {
                    "term": {
                        "TASKID": "none"
                    }
                },
                {
                    "term: {
                        "TYPE": "STOR"
                    }
                }
            ]

        which would return only those queries that have no associated TASKID and
        were sending (storing) data.

        Args:
            start (datetime.datetime): lower bound for query (inclusive)
            end (datetime.datetime): upper bound for query (exclusive)
            must (list or None): list of dictionaries to be inserted as
                additional term-level query parameters.
            scroll (bool): Use the scrolling interface if True.  If False,
                source_filter/filter_function/flush_every/flush_function are
                ignored.
        """
        if must:
            query = copy.deepcopy(QUERY)
            query['query']['constant_score']['filter']['bool']['must'] += must
        else:
            # no need to deepcopy since query_timeseries() will do that
            query = QUERY

        self.query_timeseries(query, start, end, scroll=scroll)

    def query_user(self, start, end, user):
        """Wraps query() with a user restriction

        Convenience method to constrain a query to a specific user.

        Args:
            start (datetime.datetime): lower bound for query (inclusive)
            end (datetime.datetime): upper bound for query (exclusive)
            user (str): constrain results to this username
        """
        self.query(start=start, end=end, must=[{"term": {"USER": user}}])

    def query_type(self, start, end, xfer_type):
        """Wraps query() with a type restriction

        Convenience method to constrain a query to a specific transfer type.

        Args:
            start (datetime.datetime): lower bound for query (inclusive)
            end (datetime.datetime): upper bound for query (exclusive)
            xfer_type (str): constrain results to this transfer type (STOR,
                RETR, etc).  Case sensitive.
        """
        self.query(start=start, end=end, must=[{"term": {"TYPE": xfer_type}}])

    def query_timeseries(self, query_template, start, end, scroll=True):
        """Craft and issue query that returns all overlapping records

        Args:
            query_template (dict): a query object containing at least one
                ``@timestamp`` field
            start (datetime.datetime): lower bound for query (inclusive)
            end (datetime.datetime): upper bound for query (exclusive)
            scroll (bool): Use the scrolling interface if True.  If False,
                source_filter/filter_function/flush_every/flush_function are
                ignored.
        """
        query = es.build_timeseries_query(
            query_template,
            start,
            end,
            start_key='start_date',
            end_key='end_date')

        if scroll:
            self.query_and_scroll(
                query=query,
                source_filter=SOURCE_FILTER,
                filter_function=self.filter_function,
                flush_every=self.flush_every,
                flush_function=self.flush_function)
        else:
            super(NerscGlobusLogs, self).query(query=query)

    def to_dataframe(self):
        """Converts self.scroll_pages to a DataFrame

        Returns:
            pandas.DataFrame: Contents of the last query's pages
        """
        return super(NerscGlobusLogs, self).to_dataframe(fields=SOURCE_FILTER)



#!/usr/bin/env python
"""
Load the pytokio configuration file, which is encoded as json and contains
various site-specific constants, paths, and defaults.
"""

import os
import json
from tokio.common import isstr

#: Global variable containing the configuration
CONFIG = {} 

#: Path to configuration file to load
PYTOKIO_CONFIG_FILE = "" 

#: Path of default site configuration file
DEFAULT_CONFIG_FILE = ""

def init_config():
    """Loads the global configuration.

    Loads the site-wide configuration file, then inspects relevant environment
    variables for overrides.
    """
    global CONFIG
    global PYTOKIO_CONFIG_FILE
    global DEFAULT_CONFIG_FILE

    # Escape hatch for cases where we want to load the module without initalizing
    if os.environ.get('PYTOKIO_SKIP_CONFIG') is not None:
        return

    # Set the default config path - set here rather than in the global namespace
    # so site-specific paths don't get baked into the autodoc documentation
    DEFAULT_CONFIG_FILE = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'site.json')

    # Load a pytokio config from a special location
    PYTOKIO_CONFIG_FILE = os.environ.get('PYTOKIO_CONFIG', DEFAULT_CONFIG_FILE)

    try:
        with open(PYTOKIO_CONFIG_FILE, 'r') as config_file:
            loaded_config = json.load(config_file)
    except (OSError, IOError):
        loaded_config = {}

    # Load pytokio config file and convert its keys into a set of constants
    for _key, _value in loaded_config.items():
        # config keys beginning with an underscore get skipped
        if _key.startswith('_'):
            pass

        # if setting this key will overwrite something already in the tokio.config
        # namespace, only overwrite if the existing object is something we probably
        # loaded from a json
        _old_attribute = CONFIG.get(_key.lower())
        if _old_attribute is None \
        or isstr(_old_attribute) \
        or isinstance(_old_attribute, (dict, list)):
            CONFIG[_key.lower()] = _value

    # Check for magic environment variables to override the contents of the config
    # file at runtime
    for _magic_variable in ['HDF5_FILES', 'ISDCT_FILES', 'LFSSTATUS_FULLNESS_FILES', 'LFSSTATUS_MAP_FILES', 'DARSHAN_LOG_DIRS', 'ESNET_SNMP_URI']:
        _magic_value = os.environ.get("PYTOKIO_" + _magic_variable)
        if _magic_value is not None:
            try:
                _magic_value_decoded = json.loads(_magic_value)
            except ValueError:
                CONFIG[_magic_variable.lower()] = _magic_value
            else:
                CONFIG[_magic_variable.lower()] = _magic_value_decoded

init_config()



#!/usr/bin/env python
"""Retrieve data from TOKIO Time Series files using time as inputs

Provides a mapping between dates and times and a site's time-indexed repository
of TOKIO Time Series HDF5 files.
"""

import datetime
import tokio.tools.common
import tokio.connectors.hdf5

def enumerate_h5lmts(fsname, datetime_start, datetime_end):
    """Return all time-indexed HDF5 files falling between a time range

    Given a starting datetime and (optionally) an ending datetime, return all
    HDF5 files that contain data inside of that date range (inclusive).
    """
    return tokio.tools.common.enumerate_dated_files(start=datetime_start,
                                                    end=datetime_end,
                                                    template=tokio.config.CONFIG['hdf5_files'],
                                                    lookup_key=fsname,
                                                    match_first=True)

def get_files_and_indices(fsname, dataset_name, datetime_start, datetime_end):
    """
    Given the name of an Hdf5 file and a start/end date+time, returns a list of
    tuples containing
    """
    if datetime_end is None:
        datetime_end = datetime_start
    else:
        datetime_end = datetime_end
    h5lmt_files = enumerate_h5lmts(fsname, datetime_start, datetime_end)
    output = []

    for h5lmt_file in h5lmt_files:
        with tokio.connectors.hdf5.Hdf5(h5lmt_file, mode="r") as hdf5:
            i_0 = 0
            timestamps = hdf5.get_timestamps(dataset_name)
            if datetime.datetime.fromtimestamp(timestamps[0]) <= datetime_start:
                i_0 = hdf5.get_index(dataset_name, datetime_start) # This is the first day's hdf5

            i_f = -1
            if datetime.datetime.fromtimestamp(timestamps[-1]) >= datetime_end:
                # This is the last day's hdf5
                i_f = hdf5.get_index(dataset_name, datetime_end) - 1
                # -1 because datetime_end should be exclusive
                #
                # If the last timestamp is on the first datapoint of a new day,
                # just drop the whole day to maintain exclusivity of the last
                # timestamp
                if i_f < 0:
                    continue

        output.append((h5lmt_file, i_0, i_f))
    return output

def get_dataframe_from_time_range(fsname, dataset_name, datetime_start, datetime_end):
    """
    Returns the same content as get_group_data_from_time_range into a dataframe
    """
    files_and_indices = get_files_and_indices(fsname, dataset_name, datetime_start, datetime_end)
    result = None

    if not files_and_indices:
#       raise IOError("No relevant hdf5 files found in %s between %s - %s" % (
#           tokio.config.CONFIG['h5lmt_file'],
#           datetime_start,
#           datetime_end))
        return result

    for hdf_filename in enumerate_h5lmts(fsname, datetime_start, datetime_end):
        with tokio.connectors.hdf5.Hdf5(hdf_filename, mode='r') as hdf_file:
            df_slice = hdf_file.to_dataframe(dataset_name)
            df_slice = df_slice[(df_slice.index >= datetime_start)
                                & (df_slice.index < datetime_end)]
            if result is None:
                result = df_slice
            else:
                ### append a copy--I think this is memory-inefficient
                # result = result.append(df_slice)
                # concat ?
                ### append in place--maybe more efficient than .append??
                result = result.reindex(result.index.union(df_slice.index))
                result.loc[df_slice.index] = df_slice
    return result.sort_index()



"""
Dump a lot of data out of ElasticSearch using the Python API and native
scrolling support.  Output either as native json from ElasticSearch or as
serialized TOKIO TimeSeries (TTS) HDF5 files.
"""

import os
import sys
import gzip
import json
import time
import datetime
import argparse
import warnings
import mimetypes
import collections
import multiprocessing

import dateutil.parser # because of how ElasticSearch returns time data
import dateutil.tz
import numpy

import tokio.debug
import tokio.timeseries
import tokio.connectors.collectd_es
import tokio.connectors.hdf5

SCHEMA_VERSION = "1"

DATASETS = collections.OrderedDict([
    ('datatargets/readrates', 'bytes/sec'),
    ('datatargets/writerates', 'bytes/sec'),
    ('datatargets/readoprates', 'ops/sec'),
    ('datatargets/writeoprates', 'ops/sec'),
    ('dataservers/memcached', 'bytes'),
    ('dataservers/membuffered', 'bytes'),
    ('dataservers/memfree', 'bytes'),
    ('dataservers/memused', 'bytes'),
    ('dataservers/memslab', 'bytes'),
    ('dataservers/memslab_unrecl', 'bytes'),
    ('dataservers/cpuload', '%'),
    ('dataservers/cpuuser', '%'),
    ('dataservers/cpusys', '%'),
    # these metadatasets _must_ be initialized _after_ their parent dataset
    ('dataservers/_num_cpuload', 'cpu count'),
    ('dataservers/_num_cpuuser', 'cpu count'),
    ('dataservers/_num_cpusys', 'cpu count'),
])

DATE_FMT = "%Y-%m-%dT%H:%M:%S"

# This is necessary because multiprocessing needs to be able to serialize the
# reducer that is passed back, but lambda functions are not pickleable.  So we
# pass back a string that maps to a lambda.
LAMBDA_MAPS = {
    'sum': lambda x, y: x + y,
}

def metadataset2dataset_key(metadataset_name):
    """Return the dataset name corresponding to a metadataset name

    Metadatasets are not ever stored in the HDF5 and instead are only used to
    store data needed to correctly calculate dataset values.  This function
    maps a metadataset name to its corresponding dataset name.

    Args:
        metadataset_name (str): Name of a metadataset
    Returns:
        str: Name of corresponding dataset name, or None if `metadataset_name`
        does not appear to be a metadataset name.
    """
    if '/_num_' not in metadataset_name:
        return None
    else:
        return metadataset_name.replace('/_num_', '/', 1)

def dataset2metadataset_key(dataset_key):
    """Return the metadataset name corresponding to a dataset name

    Args:
        dataset_name (str): Name of a dataset

    Returns:
        str: Name of corresponding metadataset name
    """
    return dataset_key.replace('/', '/_num_', 1)

def process_page(page):
    """
    Go through a list of docs and insert their data into a numpy matrix.  In
    the future this should be a flush function attached to the CollectdEs
    connector class.

    Args:
        page (dict): A single page of output from an Elasticsearch scroll
            query.  Should contain a ``hits`` key.
    """

    _time0 = time.time()
    inserts = []
    for doc in page:
        # basic validity checking
        if '_source' not in doc:
            warnings.warn("No _source in doc")
            print(json.dumps(doc, indent=4))
            continue
        source = doc['_source']

        # check to see if this is from plugin:disk
        if source['plugin'] == 'disk':
            # We jump through these hoops because most of TOKIO uses tz-unaware
            # local datetimes and we don't want to inject a bunch of dateutil
            # dependencies elsewhere
            # timezone-aware string
            # `-> tz-aware utc datetime
            #     `-> local datetime
            #         `-> tz-unaware local datetime
            timestamp = dateutil.parser.parse(source['@timestamp'])\
                                              .astimezone(dateutil.tz.tzlocal())\
                                              .replace(tzinfo=None)
            col_name = "%s:%s" % (source['hostname'], source['plugin_instance'])

            if source['collectd_type'] == 'disk_octets':
                val1 = source.get('read')
                val2 = source.get('write')
                if val1 is not None and val2 is not None:
                    inserts.append(('datatargets/readrates', timestamp, col_name, val1))
                    inserts.append(('datatargets/writerates', timestamp, col_name, val2))
            elif source['collectd_type'] == 'disk_ops':
                val1 = source.get('read')
                val2 = source.get('write')
                if val1 is not None and val2 is not None:
                    inserts.append(('datatargets/readoprates', timestamp, col_name, val1))
                    inserts.append(('datatargets/writeoprates', timestamp, col_name, val2))
        elif source['plugin'] == 'cpu' and 'value' in source:
            timestamp = dateutil.parser.parse(source['@timestamp'])\
                                              .astimezone(dateutil.tz.tzlocal())\
                                              .replace(tzinfo=None)
            val1 = source.get('value')
            if source['type_instance'] == 'idle':
                # note that we store (100 - idle) as load
                inserts.append(('dataservers/cpuload', timestamp, source['hostname'],
                                100.0 - val1, 'sum'))
                inserts.append(('dataservers/_num_cpuload', timestamp, source['hostname'],
                                1, 'sum'))
            elif source['type_instance'] == 'user':
                inserts.append(('dataservers/cpuuser', timestamp, source['hostname'],
                                val1, 'sum'))
                inserts.append(('dataservers/_num_cpuuser', timestamp, source['hostname'],
                                1, 'sum'))
            elif source['type_instance'] == 'system':
                inserts.append(('dataservers/cpusys', timestamp, source['hostname'],
                                val1, 'sum'))
                inserts.append(('dataservers/_num_cpusys', timestamp, source['hostname'],
                                1, 'sum'))
        elif source['plugin'] == 'memory' and 'value' in source:
            timestamp = dateutil.parser.parse(source['@timestamp'])\
                                              .astimezone(dateutil.tz.tzlocal())\
                                              .replace(tzinfo=None)
            val1 = source.get('value')
            if source['type_instance'] == 'cached':
                inserts.append(('dataservers/memcached', timestamp, source['hostname'], val1))
            elif source['type_instance'] == 'buffered':
                inserts.append(('dataservers/membuffered', timestamp, source['hostname'], val1))
            elif source['type_instance'] == 'free':
                inserts.append(('dataservers/memfree', timestamp, source['hostname'], val1))
            elif source['type_instance'] == 'used':
                inserts.append(('dataservers/memused', timestamp, source['hostname'], val1))
            elif source['type_instance'] == 'slab_recl':
                inserts.append(('dataservers/memslab', timestamp, source['hostname'], val1))
            elif source['type_instance'] == 'slab_unrecl':
                inserts.append(('dataservers/memslab_unrecl', timestamp, source['hostname'], val1))

    _timef = time.time()
    if tokio.debug.DEBUG:
        print("Extracted %d inserts in %.4f seconds" % (len(inserts), _timef - _time0))
        per_dataset = {}
        for insert in inserts:
            if insert[0] not in per_dataset:
                per_dataset[insert[0]] = 0
            per_dataset[insert[0]] += 1
        for dataset_name in sorted(per_dataset.keys()):
            print("  %6d entries for %s" % (per_dataset[dataset_name], dataset_name))
    return inserts

def update_datasets(inserts, datasets):
    """Insert list of tuples into a dataset

    Insert a list of tuples into a :class:`tokio.timeseries.TimeSeries` object serially

    Args:
        inserts (list of tuples): List of tuples which should be serially
            inserted into a dataset.  The tuples can be of the form

                * dataset name (str)
                * timestamp (:class:`datetime.datetime`)
                * column name (str)
                * value

            or

                * dataset name (str)
                * timestamp (:class:`datetime.datetime`)
                * column name (str)
                * value
                * reducer name (str)

            where

                * `dataset name` is the key used to retrieve a target
                  :class:`tokio.timeseries.TimeSeries` object from the `datasets`
                  argument
                * `timestamp` and `column name` reference the element to be udpated
                * value is the new value to insert into the given (`timestamp`,
                  `column name`) location within `dataset`.
                * `reducer name` is None (to just replace whatever value
                  currently exists in the (`timestamp`, `column name`) location,
                  or 'sum' to add `value` to the existing value.

        datasets (dict): Dictionary mapping dataset names (str) to
            :class:`tokio.timeseries.TimeSeries` objects

    Returns:
        int: number of elements in `inserts` which were not inserted because
        their timestamp value was out of the range of the dataset to be updated.
    """
    data_volume = {}
    errors = {}
    for key in datasets:
        data_volume[key] = 0.0
        errors[key] = 0

    for insert in inserts:
        try:
            if len(insert) == 4:
                (dataset_name, timestamp, col_name, value) = insert
                reducer = None
            else:
                (dataset_name, timestamp, col_name, value, reducer_name) = insert
                reducer = LAMBDA_MAPS.get(reducer_name)
        except ValueError:
            print(insert)
            raise

        if datasets[dataset_name].insert_element(timestamp, col_name, value, reducer):
            data_volume[dataset_name] += value
            tidx, cidx = datasets[dataset_name].get_insert_pos(timestamp, col_name)
        else:
            errors[dataset_name] += 1

    # Update dataset metadata
    for key in datasets:
        unit = DATASETS.get(key, "unknown")
        datasets[key].dataset_metadata.update({'version': SCHEMA_VERSION, 'units': unit})
        datasets[key].group_metadata.update({'source': 'collectd_disk'})

    index_errors = sum(errors.values())
    if index_errors > 0:
        warnings.warn("Out-of-bounds indices (%d total) were detected" % index_errors)

    if tokio.debug.DEBUG:
        for key in data_volume:
            data_volume[key] *= datasets[key].timestep
        update_str = "Added "
        update_str = "Added %(datatargets/readrates)d bytes/" \
            + "%(datatargets/readoprates)d ops of read, " \
            + "%(datatargets/writerates)d bytes/" \
            + "%(datatargets/writeoprates)d ops of write, " \
            + "%(dataservers/cpuload).1f%% of cpu load-seconds, " \
            + "%(dataservers/memused)d bytes of memory-seconds"
        print(update_str % data_volume)
    return index_errors

def reset_timeseries(timeseries, start, end, value=-0.0):
    """Zero out a region of a tokio.timeseries.TimeSeries dataset

    Args:
        timeseries (tokio.timeseries.TimeSeries): data from a subset should be zeroed
        start (datetime.datetime): Time at which zeroing of all columns in
            `timeseries` should begin
        end (datetime.datetime): Time at which zeroing all columns in
            `timeseries` should end (exclusive)
        value: value which should be set in every element being reset

    Returns:
        Nothing
    """
    index0, _ = timeseries.get_insert_pos(start, None)
    indexf, _ = timeseries.get_insert_pos(end, None)
    timeseries.dataset[index0:indexf, :] = value

def normalize_cpu_datasets(inserts, datasets):
    """Normalize CPU load datasets

    Divide each element of CPU datasets by the number of CPUs counted at each
    point in time.  Necessary because these measurements are reported on a
    per-core basis, but not all cores may be reported for each timestamp.

    Args:
        inserts (list of tuples): list of inserts that were used to populate
            datasets
        datasets (dict of TimeSeries): all of the datasets being populated

    Returns:
        Nothing
    """
    dataset_names = set(['dataservers/cpuload', 'dataservers/cpuuser', 'dataservers/cpusys'])
    num_dataset_names = {}

    norm_elements = {}
    for dataset_name in dataset_names:
        norm_elements[dataset_name] = set([])
        num_dataset_names[dataset_name] = dataset2metadataset_key(dataset_name)

    # build a set of all elements that must be divided
    for insert in inserts:
        (dataset_name, timestamp, col_name) = insert[0:3]
        if dataset_name in dataset_names:
            # get the position of this element to be inserted
            t_index, c_index = datasets[dataset_name].get_insert_pos(timestamp, col_name)
            if t_index is not None and c_index is not None:
                norm_elements[dataset_name].add((t_index, c_index))

    # now divide each element to be divided
    for dataset_name in dataset_names:
        num_dataset_name = num_dataset_names[dataset_name]
        for t_index, c_index in norm_elements[dataset_name]:
            datasets[dataset_name].dataset[t_index, c_index] /= \
                datasets[num_dataset_name].dataset[t_index, c_index]
        # convert NaNs (0.0 / 0.0) back to -0.0
        datasets[dataset_name].dataset[numpy.isnan(datasets[dataset_name].dataset)] = -0.0

def pages_to_hdf5(pages, output_file, init_start, init_end, query_start, query_end,
                  timestep, num_servers, devices_per_server, threads=1):
    """Stores a page from Elasticsearch query in an HDF5 file
    Take pages from ElasticSearch query and store them in output_file

    Args:
        pages (list): A list of page objects (dictionaries)
        output_file (str): Path to an HDF5 file in which page data should be
            stored
        init_start (datetime.datetime): Lower bound of time (inclusive) to be
            stored in the ``output_file``.  Used when creating a non-existent
            HDF5 file.
        init_end (datetime.datetime): Upper bound of time (inclusive) to be
            stored in the ``output_file``.  Used when creating a non-existent
            HDF5 file.
        query_start (datetime.datetime): Retrieve data greater than or equal to
            this time from Elasticsearch
        query_end (datetime.datetime); Retrieve data less than this time from
            Elasticsearch
        timestep (int): Time, in seconds, between successive sample intervals
            to be used when initializing ``output_file``
        num_servers (int): Number of discrete servers in the cluster.  Used
            when initializing ``output_file``.
        devices_per_server (int): Number of SSDs per server.  Used when
            initializing ``output_file``.
        threads (int): Number of parallel threads to utilize when parsing the
            Elasticsearch output
    """
    datasets = {}

    file_exists = False
    if os.path.isfile(output_file):
        file_exists = True

    with tokio.connectors.hdf5.Hdf5(output_file) as hdf5_file:
        schema_version = hdf5_file.get_version()

        # New files have a blank slate and should use the latest; existing files may
        # have orphaned duplicate data if two schemata are encoded at once
        if file_exists and schema_version != SCHEMA_VERSION:
            warnings.warn("%s existing version %s will be upgraded in-place to %s"
                % (output_file, schema_version, SCHEMA_VERSION))

        # Update to the latest schema version no matter what
        schema_version = SCHEMA_VERSION
        hdf5_file.attrs['version'] = SCHEMA_VERSION
        schema = tokio.connectors.hdf5.SCHEMA.get(schema_version)
        if schema is None:
            raise KeyError("Schema version %d in %s is not known by connectors.hdf5" % (SCHEMA_VERSION, output_file))

        # Initialize datasets
        for dataset_name in DATASETS:
            hdf5_dataset_name = schema.get(dataset_name)
            if hdf5_dataset_name is None:
                if '/_' not in dataset_name:
                    warnings.warn("Dataset %s in %s is not in schema version %s (passing through)" % (dataset_name, output_file, schema_version))
                hdf5_dataset_name = dataset_name
            if dataset_name.lstrip('/').startswith('datatargets'):
                num_columns = num_servers * devices_per_server
            else:
                num_columns = num_servers

            # If this is a metadataset, initialize it with the shape and columns of
            # the dataset it describes so that we can use the same index values for
            # both.  This requires the parent dataset's TimeSeries to already be
            # defined and attached, which requires DATASETS to be an OrderedDict so
            # that we aren't initializing metadatasets before datasets.
            real_dataset_name = metadataset2dataset_key(dataset_name)
            if real_dataset_name:
                if real_dataset_name not in datasets:
                    raise KeyError("Cannot init metadataset %s; dataset %s does not exist" %
                                   (dataset_name, real_dataset_name))
                global_start = datetime.datetime.fromtimestamp(datasets[real_dataset_name].timestamps[0])
                global_end = datetime.datetime.fromtimestamp(datasets[real_dataset_name].timestamps[-1] + datasets[real_dataset_name].timestep)
                timeseries = tokio.timeseries.TimeSeries(dataset_name=hdf5_dataset_name,
                                                         start=global_start,
                                                         end=global_end,
                                                         timestep=timestep,
                                                         num_columns=num_columns,
                                                         column_names=datasets[real_dataset_name].columns)

                # we can't update an average, so zero out all prior values that
                # we will be overwriting
                reset_timeseries(datasets[real_dataset_name], query_start, query_end)
            else:
                timeseries = hdf5_file.to_timeseries(dataset_name=hdf5_dataset_name)
                if timeseries is None:
                    timeseries = tokio.timeseries.TimeSeries(dataset_name=hdf5_dataset_name,
                                                             start=init_start,
                                                             end=init_end,
                                                             timestep=timestep,
                                                             num_columns=num_columns)
            datasets[dataset_name] = timeseries

        # Process all pages retrieved (this is computationally expensive)
        _time0 = time.time()
        updates = []
        if threads > 1:
            pool = multiprocessing.Pool(threads)
            for update in pool.imap_unordered(process_page, pages):
                updates.append(update)
            # explicitly terminate to prevent HDF5 locking problems caused by
            # un-gc'ed file handles
            pool.terminate()
        else:
            for page in pages:
                updates.append(process_page(page))
        _timef = time.time()
        _extract_time = _timef - _time0
        tokio.debug.debug_print("Extracted %d elements from %d pages in %.4f seconds" \
                                % (sum([len(x) for x in updates]),
                                   len(pages),
                                   _extract_time))

        # Take the processed list of data to insert and actually insert them
        _time0 = time.time()
        for update in updates:
            update_datasets(update, datasets)
        _timef = time.time()
        _update_time = _timef - _time0
        if tokio.debug.DEBUG:
            print("Inserted %d elements from %d pages in %.4f seconds" \
                % (len(updates), len(pages), _update_time))
            print("Processed %d pages in %.4f seconds" \
                % (len(pages), _extract_time + _update_time))

        for update in updates:
            normalize_cpu_datasets(update, datasets)

        # Write datasets out to HDF5 file
        _time0 = time.time()
        for dataset_name, dataset in datasets.items():
            if '/_' not in dataset_name:
                hdf5_file.commit_timeseries(dataset)

    if tokio.debug.DEBUG:
        print("Committed data to disk in %.4f seconds" % (time.time() - _time0))

def main(argv=None):
    """Entry point for the CLI interface
    """
    warnings.simplefilter('always', UserWarning) # One warning per invalid file

    # Parse CLI options
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("query_start", type=str,
                        help="start time of query in %s format" % DATE_FMT)
    parser.add_argument("query_end", type=str,
                        help="end time of query in %s format" % DATE_FMT)
    parser.add_argument('--init-start', type=str, default=None,
                        help='min timestamp if creating new output file, in %s format' % DATE_FMT
                        + ' (default: same as start)')
    parser.add_argument('--init-end', type=str, default=None,
                        help='max timestamp if creating new output file, in %s format' % DATE_FMT
                        + ' (default: same as end)')
    parser.add_argument('--debug', action='store_true',
                        help="produce debug messages")
    parser.add_argument('--num-nodes', type=int, default=288,
                        help='number of expected burst buffer nodes (default: 288)')
    parser.add_argument('--ssds-per-node', type=int, default=4,
                        help='number of SSDs in each BB node (default: 4)')
    parser.add_argument('--timestep', type=int, default=10,
                        help='collection frequency, in seconds (default: 10)')
    parser.add_argument('--timeout', type=int, default=30,
                        help='ElasticSearch timeout time (default: 30)')
    parser.add_argument('--threads', type=int, default=1,
                        help='parallel threads for document extraction (default: 1)')
    parser.add_argument('--input', type=str, default=None,
                        help="use cached ElasticSearch json as input")
    parser.add_argument("-o", "--output", type=str, default='output.hdf5',
                        help="output file (default: output.hdf5)")
    parser.add_argument('-h', '--host', type=str, default="localhost",
                        help="hostname of ElasticSearch endpoint (default: localhost)")
    parser.add_argument('-p', '--port', type=int, default=9200,
                        help="port of ElasticSearch endpoint (default: 9200)")
    parser.add_argument('-i', '--index', type=str, default='cori-collectd-*',
                        help='ElasticSearch index to query (default:cori-collectd-*)')
    args = parser.parse_args(argv)

    if args.debug:
        tokio.debug.DEBUG = True

    # Convert CLI options into datetime
    try:
        query_start = datetime.datetime.strptime(args.query_start, DATE_FMT)
        query_end = datetime.datetime.strptime(args.query_end, DATE_FMT)
        init_start = query_start
        init_end = query_end
        if args.init_start:
            init_start = datetime.datetime.strptime(args.init_start, DATE_FMT)
        if args.init_end:
            init_end = datetime.datetime.strptime(args.init_end, DATE_FMT)
    except ValueError:
        sys.stderr.write("Start and end times must be in format %s\n" % DATE_FMT)
        raise

    # Basic input bounds checking
    if query_start >= query_end:
        raise Exception('query_start >= query_end')
    elif init_start >= init_end:
        raise Exception('init_start >= init_end')
    elif args.timestep < 1:
        raise Exception('--timestep must be > 0')

    # Read input from a cached json file (generated previously via the --json
    # option) or by querying ElasticSearch?
    if args.input is None:
        ### Try to connect
        esdb = tokio.connectors.collectd_es.CollectdEs(
            host=args.host,
            port=args.port,
            index=args.index,
            timeout=args.timeout)

        pages = None
        for plugin_query in [tokio.connectors.collectd_es.QUERY_CPU_DATA,
                             tokio.connectors.collectd_es.QUERY_DISK_DATA,
                             tokio.connectors.collectd_es.QUERY_MEMORY_DATA]:
            esdb.query_timeseries(plugin_query,
                                  query_start,
                                  query_end)
            if pages is None:
                pages = esdb.scroll_pages
            else:
                pages += esdb.scroll_pages

            tokio.debug.debug_print("Loaded results from %s:%s" % (args.host, args.port))
            pages_to_hdf5(pages=pages,
                          output_file=args.output,
                          init_start=init_start,
                          init_end=init_end,
                          query_start=query_start,
                          query_end=query_end,
                          timestep=args.timestep,
                          num_servers=args.num_nodes,
                          devices_per_server=args.ssds_per_node,
                          threads=args.threads)
    else:
        _, encoding = mimetypes.guess_type(args.input)
        if encoding == 'gzip':
            input_file = gzip.open(args.input, 'r')
        else:
            input_file = open(args.input, 'r')
        pages = json.load(input_file)
        input_file.close()
        tokio.debug.debug_print("Loaded results from %s" % args.input)
        pages_to_hdf5(pages=pages,
                      output_file=args.output,
                      init_start=init_start,
                      init_end=init_end,
                      query_start=query_start,
                      query_end=query_end,
                      timestep=args.timestep,
                      num_servers=args.num_nodes,
                      devices_per_server=args.ssds_per_node,
                      threads=args.threads)

    print("Wrote output to %s" % args.output)



#!/usr/bin/env python
"""Connect to Darshan logs.

This connector provides an interface into Darshan logs created by Darshan 3.0 or
higher and represents the counters and data contained therein as a Python
dictionary.  This dictionary has the following structure, where ``block``
denote literal key names.

* ``header`` which contains key-value pairs corresponding to each line in the
  header.  ``exe`` and ``metadata`` are lists; the other keys correspond to a
  single scalar value.

  * ``compression``, ``end_time``, ``end_time_string``, ``exe``, etc

* ``counters``

  * `modulename` which is ``posix``, ``lustre``, ``stdio``, etc

    * `recordname`, which is usually the full path to a file opened by the
      profiled application _or_ ``_perf`` (contains performance summary metrics)
      or ``_total`` (contains aggregate file statistics)

      * ranknum which is a string (``0``, ``1``, etc or ``-1``)

        * `counternames`, which depends on the Darshan module defined by
          `modulename` above

* ``mounts`` which is the mount table with keys of a path to a mount location
  and values of the file system type

The `counternames` are module-specific and have their module name prefix
stripped off.  The following counter names are examples of what a Darshan log
may expose through this connector for the ``posix`` module:

* ``BYTES_READ`` and ``BYTES_WRITTEN`` - number of bytes read/written to the file
* ``MAX_BYTE_WRITTEN`` and ``MAX_BYTE_READ`` - highest byte written/read; useful if an application re-reads or re-writes a lot of data
* ``WRITES`` and ``READS`` - number of write and read ops issued
* ``F_WRITE_TIME`` and ``F_READ_TIME`` - amount of time spent inside write and read calls (in seconds)
* ``F_META_TIME`` - amount of time spent in metadata (i.e., non-read/write) calls

Similarly the ``lustre`` module provides the following counter keys:

* ``MDTS`` - number of MDTs in the underlying file system
* ``OSTS`` - number of OSTs in the underlying file system
* ``OST_ID_0`` - the OBD index for the 0th OST over which the file is striped
* ``STRIPE_OFFSET`` - the setting used to define stripe offset when the file was created
* ``STRIPE_SIZE`` - the size, in bytes, of each stripe
* ``STRIPE_WIDTH`` - how many OSTs the file touches

Note:
    This connector presently relies on ``darshan-parser`` to convert the binary
    logs to ASCII, then convert the ASCII into Python objects.  In the future,
    we plan on using the Python API provided by darshan-utils to circumvent the
    ASCII translation.
"""

import os
import re
import json
from tokio.connectors.common import SubprocessOutputDict

DARSHAN_PARSER_BIN = 'darshan-parser'

DARSHAN_FILENAME_REX = re.compile(r'([^_%s]+)_([^%s]*?)_id(\d+)_(\d+)-(\d+)-(\d+)-(\d+)_(\d+).darshan' % (os.path.sep, os.path.sep))

class Darshan(SubprocessOutputDict):
    def __init__(self, log_file=None, *args, **kwargs):
        """Initialize the object from either a Darshan log or a cache file.

        Configures the object's internal state to operate on a Darshan
        log file or a cached JSON representation of a previously processed
        Darshan log.

        Args:
            log_file (str, optional): Path to a Darshan log to be processed
            cache_file (str, optional): Path to a Darshan log's contents cached
            *args: Passed to tokio.connectors.common.SubprocessOutputDict
            *kwargs: Passed to tokio.connectors.common.SubprocessOutputDict

        Attributes:
            log_file (str): Path to the Darshan log file to load
        """
        super(Darshan, self).__init__(*args, **kwargs)
        self.log_file = log_file
        self._parser_mode = None
        self.subprocess_cmd = [DARSHAN_PARSER_BIN]
        self.filename_metadata = {}
        if log_file is None:
            self.load()
        else:
            # try to infer metadata from file name
            regex_match = DARSHAN_FILENAME_REX.search(log_file)
            if regex_match:
                self.filename_metadata['username'] = regex_match.group(1)
                self.filename_metadata['exename'] = regex_match.group(2)
                self.filename_metadata['jobid'] = regex_match.group(3)
                self.filename_metadata['start_month'] = int(regex_match.group(4))
                self.filename_metadata['start_day'] = int(regex_match.group(5))
                self.filename_metadata['start_second_in_day'] = int(regex_match.group(6))
                self.filename_metadata['logmod'] = int(regex_match.group(7))
                self.filename_metadata['shutdown_secs'] = int(regex_match.group(8))

    def __repr__(self):
        """Serialize self into JSON.

        Returns:
            str: JSON representation of the object
        """
        return json.dumps(list(self.values()))

    def load(self):
        if self.from_string is not None:
            self.load_str(self.from_string)
        elif self.cache_file:
            self.load_cache()
        elif self.log_file is None:
            raise Exception("parameters should be provided (at least log_file or cache_file)")

    def darshan_parser_base(self):
        """Populate data produced by ``darshan-parser --base``

        Runs the ``darshan-parser --base`` and convert all results into
        key-value pairs which are inserted into the object.

        Returns:
            dict: Dictionary containing all key-value pairs generated by running
            ``darshan-parser --base``.  These values are also accessible via the
            `BASE` key in the object.
        """
        self._parser_mode = "BASE"
        return self._darshan_parser()

    def darshan_parser_total(self):
        """Populate data produced by ``darshan-parser --total``

        Runs the ``darshan-parser --total`` and convert all results into
        key-value pairs which are inserted into the object.

        Returns:
            dict: Dictionary containing all key-value pairs generated by running
            ``darshan-parser --total``.  These values are also accessible via
            the `TOTAL` key in the object.
        """
        self._parser_mode = "TOTAL"
        return self._darshan_parser()

    def darshan_parser_perf(self):
        """Populate data produced by ``darshan-parser --perf``

        Runs the ``darshan-parser --perf`` and convert all results into
        key-value pairs which are inserted into the object.

        Returns:
            dict: Dictionary containing all key-value pairs generated by running
            ``darshan-parser --perf``.  These values are also accessible via the
            `PERF` key in the object.
        """
        self._parser_mode = "PERF"
        return self._darshan_parser()

    def _darshan_parser(self):
        """Call darshan-parser to initialize values in self
        """

        if self.log_file is None:
            return self

        if self._parser_mode in ["BASE", "TOTAL", "PERF"]:
            darshan_flag = "--" + self._parser_mode.lower()
        else:
            self._parser_mode = "BASE"
            darshan_flag = ""

        args = [darshan_flag, self.log_file]

        self._load_subprocess(*args)

        return self

    def load_str(self, input_str):
        """Load from either a json cache or the output of darshan-parser

        Args:
            input_str (str): Stdout of the darshan-parser command
        """
        loaded_data = None
        try:
            loaded_data = json.loads(input_str)
        except ValueError:
            pass

        if loaded_data:
            self.__setitem__(loaded_data)
        else:
            self._parse_darshan_parser(input_str)

    def _parse_darshan_parser(self, output_str):
        """Load values from output of darshan-parser
        """
        def is_valid_counter(counter):
            """
            if counter is not None, this line is valid (return True)
            if counter is None but we can identify a module section, return it
            if counter is None but we cannot identify a module section, return False
            """
            if counter is None:
                match = module_rex.search(line)
                if match is not None:
                    module_section = match.group(1)
                    module_section = module_section.replace('-', '') # because of "MPI-IO" and "MPIIO"
                    module_section = module_section.replace('/', '') # because of "BG/Q" and "BGQ"
                    return False, module_section
                return False, None
            return True, None

        def insert_record(section, module, file_name, rank, counter, value, counter_prefix=None):
            """
            Embed a counter=value pair deep within the darshan_data structure based
            on a bunch of nested keys.
            """
            # Force the local shadow of 'module' to lowercase
            module = module.lower()
            # Assert that the counter actually belongs to the current module
            if counter_prefix is not None:
                if counter.startswith(counter_prefix):
                    # Strip off the counter_prefix from the counter name
                    counter = counter[len(counter_prefix):]
                else:
                    raise Exception("counter %s does not start with prefix %s" % (counter, counter_prefix))

            # Otherwise insert the record--this logic should be made more flexible
            if section not in self:
                self[section] = {}
            if module not in self[section]:
                self[section][module] = {}
            if file_name not in self[section][module]:
                self[section][module][file_name] = {}
            if rank is None:
                insert_base = self[section][module][file_name]
            else:
                if rank not in self[section][module][file_name]:
                    self[section][module][file_name][rank] = {}
                insert_base = self[section][module][file_name][rank]

            if counter in insert_base:
                raise Exception("Duplicate counter %s found in %s->%s->%s (rank=%s)" % (counter, section, module, file_name, rank))
            else:
                if '.' in value:
                    value = float(value)
                else:
                    value = int(value)
                insert_base[counter] = value

        section = None
        counter = None
        module_section = None
        # This regex must match every possible module name
        module_rex = re.compile(r'^# ([A-Z\-0-9/]+) module data\s*$')

        for line in output_str.splitlines():
            # Is this the start of a new section?
            # Why do we look at section, refactorize failed
            if section is None and line.startswith("# darshan log version:"):
                section = "header"
                if section not in list(self.keys()):
                    self[section] = {}

            elif section == "header" and line.startswith("# mounted file systems"):
                section = "mounts"
                if section not in list(self.keys()):
                    self[section] = {}

            elif section == "mounts" and line.startswith("# **********************"):  # understand the utility of these stars
                section = "counters"
                if section not in list(self.keys()):
                    self[section] = {}

            # otherwise use the appropriate parser for this section
            if section == "header":
                key, val = parse_header(line)
                if key is None:
                    pass
                elif key == "metadata":
                    if key not in self[section]:
                        self[section][key] = []
                    self[section][key].append(val)
                else:
                    self[section][key] = val
            elif section == 'mounts':
                key, val = parse_mounts(line)
                if key is not None:
                    self[section][key] = val

            elif section == 'counters':
                if self._parser_mode == "BASE":
                    # module, rank, record_id, counter, value, file_name, mount_pt, fs_type = parse_base_counters(line)
                    _, rank, _, counter, value, file_name, _, _ = parse_base_counters(line)
                    if module_section is not None:
                        # If it is none, is_valid_counter check below will bail
                        counter_prefix = module_section + "_"
                elif self._parser_mode == "TOTAL":
                    counter, value = parse_total_counters(line)
                    file_name = '_total'
                    rank = None
                    if module_section is not None:
                        counter_prefix = 'total_%s_' % module_section
                elif self._parser_mode == "PERF":
                    counter, value = parse_perf_counters(line)
                    file_name = '_perf'
                    rank = None
                    counter_prefix = None
                else:
                    continue

            # If no valid counter found, is this the start of a new module?
            valid, new_module_section = is_valid_counter(counter)
            if new_module_section is not None:
                module_section = new_module_section
            if valid is False:
                continue
            # Reminder: section is already defined as the global parser state
            insert_record(section=section,
                          module=module_section,
                          file_name=file_name,
                          rank=rank,
                          counter=counter,
                          value=value,
                          counter_prefix=counter_prefix)
        return self

def parse_header(line):
    """Parse the header lines of ``darshan-parser``.

    Accepts a line that may or may not be a header line as printed by
    ``darshan-parser``.  Such header lines take the form::

    # darshan log version: 3.10
    # compression method: ZLIB
    # exe: /home/user/bin/myjob.exe --whatever
    # uid: 69615

    If it is a valid header line, return a key-value pair corresponding to
    its decoded contents.

    Args:
        line (str): A single line of output from ``darshan-parser``

    Returns:
        tuple: Returns a (key, value) corresponding to the key and value
        decoded from the header `line`, or ``(None, None)`` if the line does
        not appear to contain a known header field.
        """

    if line.startswith("# darshan log version:"):
        return 'version', line.split()[-1]
    elif line.startswith("# compression method:"):
        return "compression", line.split()[-1]
    elif line.startswith("# exe:"):
        return "exe", line.split()[2:]
    elif line.startswith("# uid:"):
        return "uid", int(line.split()[-1])
    elif line.startswith("# jobid:"):
        return 'jobid', line.split()[-1]
    elif line.startswith("# start_time:"):
        return 'start_time', int(line.split()[2])
    elif line.startswith("# start_time_asci:"):
        return 'start_time_string', line.split(None, 2)[-1].strip()
    elif line.startswith("# end_time:"):
        return 'end_time', int(line.split()[2])
    elif line.startswith("# end_time_asci:"):
        return 'end_time_string', line.split(None, 2)[-1].strip()
    elif line.startswith("# nprocs:"):
        return "nprocs", int(line.split()[-1])
    elif line.startswith("# run time:"):
        return "walltime", int(line.split()[-1])
    elif line.startswith("# metadata:"):
        return "metadata", line.split(None, 2)[-1].strip()
    return None, None


def parse_mounts(line):
    """Parse a mount table line from ``darshan-parser``.

    Accepts a line that may or may not be a mount table entry from
    ``darshan-parser``.  Such lines take the form::

    # mount entry:  /usr/lib64/libibverbs.so.1.0.0  dvs

    If `line` is a valid mount table entry, return a key-value
    representation of its contents.

    Args:
        line (str): A single line of output from ``darshan-parser``

    Returns:
        tuple: Returns a (key, value) corresponding to the mount table
        entry, or ``(None, None)`` if the line is not a valid mount
        table entry.
    """
    if line.startswith("# mount entry:\t"):
        key, val = line.split('\t')[1:3]
        return key, val.strip()
    return None, None

def parse_base_counters(line):
    """Parse a counter line from ``darshan-parser --base``.

    Parse the line containing an actual counter's data.  It is a tab-delimited
    line of the form

    module, rank, record_id, counter, value, file_name, mount_pt, fs_type

    Args:
        line (str): A single line of output from ``darshan-parser --base``

    Returns:
        tuple: Returns a tuple containing eight values, each corresponding
        to a field represented in `line`.  If `line` is not a valid counter
        line, all values will be None.
    """
    if not line.startswith("#"):
        args = line.split('\t')
        if len(args) == 8:
            return tuple(args)
    return None, None, None, None, None, None, None, None

def parse_total_counters(line):
    """Parse a counter line from ``darshan-parser --total``.

    Parse a line containing counter data from ``darshan-parser --total``.
    Such lines are of the form:

        total_MPIIO_F_READ_END_TIMESTAMP: 0.000000

    Args:
        line (str): A single line of output from ``darshan-parser --total``

    Returns:
        tuple: Returns a single (key, value) pair corresponding to a
        counted metric and its total value.  If `line` is not a valid
        counter line, ``(None, None)`` are returned.
    """
    if not line.startswith("#"):
        args = line.split(':')
        if len(args) == 2:
            return args[0].strip(), args[1].strip()
    return None, None

def parse_perf_counters(line):
    """Parse a counter line from ``darshan-parser --perf``.

    Parse a line containing counter data from ``darshan-parser --perf``.
    Such lines look like::

        # total_bytes: 2199023259968
        # unique files: slowest_rank_io_time: 0.000000
        # shared files: time_by_cumul_io_only: 39.992327
        # agg_perf_by_slowest: 28670.996545

    Args:
        line (str): A single line of output from ``darshan-parser --perf``

    Returns:
        tuple: Returns a single (key, value) pair corresponding to the
        performance metric encoded in `line`.  If `line` is not a valid
        performance counter line, ``(None, None)`` is returned.
    """
    if line.startswith('# total_bytes:') or line.startswith('# agg_perf_by'):
        key, value = line[2:].split(':')
    elif line.startswith('# unique files:') or line.startswith('# shared files:'):
        key_suffix, key, value = line[2:].split(':')
        key += '_%s' % key_suffix.replace(' ', '_')
    else:
        return None, None

    return key.strip(), value.strip()



"""
Provide a CLI interface for :meth:`tokio.connectors.esnet_snmp.EsnetSnmp.to_dataframe`
and :meth:`tokio.connectors.esnet_snmp.EsnetSnmp.save_cache` methods.
"""

import sys
import datetime
import argparse
import tokio.connectors.esnet_snmp
import tokio.config

DATE_FMT = "%Y-%m-%dT%H:%M:%S"
DATE_FMT_PRINT = "YYYY-MM-DDTHH:MM:SS"

def main(argv=None):
    """Entry point for the CLI interface
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("-s", "--start", type=str,
                        help="start time of query in %s format" % DATE_FMT_PRINT)
    parser.add_argument("-e", "--end", type=str,
                        help="end time of query in %s format" % DATE_FMT_PRINT)
    parser.add_argument("-j", "--json", action="store_true", default=True,
                        help="return output in JSON format")
    parser.add_argument("-c", "--csv", action="store_true",
                        help="return output in CSV format")
#   parser.add_argument('--debug', action='store_true',
#                       help="produce debug messages")
    parser.add_argument("-i", "--input", type=str, default=None,
                        help="read input from this JSON instead of accessing REST API")
    parser.add_argument("-o", "--output", type=str, default=None,
                        help="output file")
    parser.add_argument("endpoints", type=str,
                        help='endpoint(s) to process, either a center name'
                        + ' ("nersc") or comma-separated list of format'
                        + ' "endpoint0:if0,endpoint1:if1,..." etc')
    args = parser.parse_args(argv)

#   if args.debug:
#       tokio.DEBUG = True

    # Parse endpoints and interfaces
    if ':' not in args.endpoints:
        query_args = tokio.config.CONFIG.get('esnet_snmp_interfaces', {}).get(args.endpoints, {}).copy()
    else:
        query_args = {}
        for kvpair in args.endpoints.split(","):
            key, value = kvpair.split(":", 1)
            if key not in query_args:
                query_args[key] = []
            query_args[key].append(value)

    if not query_args:
        errstr = "Invalid endpoint specification.\n\n"
        errstr += "Valid endpoints:" + "\n  ".join(tokio.config.CONFIG.get('esnet_snmp_interfaces', {}).keys())
        errstr += "\n\nor endpoint:interface[,endpoint:interface[,endpoint:interface]]"
        sys.stderr.write(errstr + "\n")
        raise ValueError("Invalid endpoint specification")

    if args.end and not args.start:
        parser.error("--start must be specified with --end")
    try:
        if args.start:
            start = datetime.datetime.strptime(args.start, DATE_FMT)
        if args.end:
            end = datetime.datetime.strptime(args.end, DATE_FMT)
    except ValueError:
        raise ValueError("Start and end times must be in format %s\n" % DATE_FMT_PRINT)

    if not args.start:
        start = datetime.datetime.now() - datetime.timedelta(hours=1)
    if not args.end:
        end = start + datetime.timedelta(hours=1)

    # Basic input bounds checking
    if start >= end:
        raise ValueError('--start >= --end')

    if args.input:
        esnetdata = tokio.connectors.esnet_snmp.EsnetSnmp(start=start, end=end, input_file=args.input)
    else:
        # Initialize query interface
        esnetdata = tokio.connectors.esnet_snmp.EsnetSnmp(start=start, end=end)

        # Query each endpoint:interface specified
        for endpoint, interfaces in query_args.items():
            for interface in interfaces:
                ret = esnetdata.get_interface_counters(endpoint=endpoint, interface=interface,
                                                       direction="in",
                                                       agg_func='average')
                ret = esnetdata.get_interface_counters(endpoint=endpoint,
                                                       interface=interface,
                                                       direction="out",
                                                       agg_func='average')

    # Serialize the object
    cache_file = args.output
    if cache_file is not None:
        print("Caching to %s" % cache_file)

    if args.csv:
        if cache_file is None:
            print(esnetdata.to_dataframe().to_csv())
        else:
            esnetdata.to_dataframe().to_csv(cache_file)
    elif args.json:
        if cache_file is not None:
            esnetdata.save_cache(cache_file)
        else:
            esnetdata.save_cache(cache_file, indent=4, sort_keys=True)
            sys.stdout.write("\n")
#   else:
#       # should never be encountered; default is args.json = True



"""Command-line interface into the nersc_globuslogs connector
"""

import sys
import gzip
import json
import time
import datetime
import argparse
import warnings
import mimetypes

import tokio.debug
import tokio.connectors.nersc_globuslogs

DATE_FMT = "%Y-%m-%dT%H:%M:%S"

def main(argv=None):
    """Entry point for the CLI interface
    """
    warnings.simplefilter('always', UserWarning) # One warning per invalid file

    # Parse CLI options
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("query_start", type=str,
                        help="start time of query in %s format" % DATE_FMT)
    parser.add_argument("query_end", type=str,
                        help="end time of query in %s format" % DATE_FMT)
    parser.add_argument('--debug', action='store_true',
                        help="produce debug messages")
    parser.add_argument('--timeout', type=int, default=30,
                        help='ElasticSearch timeout time (default: 30)')
    parser.add_argument('--input', type=str, default=None,
                        help="use cached output from previous ES query")
    parser.add_argument("-o", "--output", type=str, default=None,
                        help="output file")
    parser.add_argument('-h', '--host', type=str, default="localhost",
                        help="hostname of ElasticSearch endpoint (default: localhost)")
    parser.add_argument('-p', '--port', type=int, default=9200,
                        help="port of ElasticSearch endpoint (default: 9200)")
    parser.add_argument('-i', '--index', type=str, default='dtn-dtn-log*',
                        help='ElasticSearch index to query (default:dtn-dtn-log*)')
    parser.add_argument("-c", "--csv", action="store_true", help="return output in CSV format")
    parser.add_argument('--user', type=str, default=None,
                        help='limit results to transfers owned by this user')
    parser.add_argument('--type', type=str, default=None,
                        help='limit results to either STOR, RETR, or NLST')
    args = parser.parse_args(argv)

    if args.debug:
        tokio.DEBUG = True

    # Convert CLI options into datetime
    try:
        query_start = datetime.datetime.strptime(args.query_start, DATE_FMT)
        query_end = datetime.datetime.strptime(args.query_end, DATE_FMT)
    except ValueError:
        sys.stderr.write("Start and end times must be in format %s\n" % DATE_FMT)
        raise

    # Basic input bounds checking
    if query_start >= query_end:
        raise Exception('query_start >= query_end')

    # Read input from a cached json file (generated previously) or by querying
    # Elasticsearch directly
    must = []
    if args.input is None:
        ### Try to connect
        esdb = tokio.connectors.nersc_globuslogs.NerscGlobusLogs(
            host=args.host,
            port=args.port,
            index=args.index,
            timeout=args.timeout)

        if args.user:
            must.append({"term": {"USER": args.user}})
        if args.type:
            must.append({"term": {"TYPE": args.type.upper()}})

        tokio.debug.debug_print("Loaded results from %s:%s" % (args.host, args.port))
    else:
        esdb = tokio.connectors.nersc_globuslogs.NerscGlobusLogs.from_cache(args.input)
        tokio.debug.debug_print("Loaded results from %s" % args.input)

    esdb.query(query_start, query_end, must=must)

    # Write output
    cache_file = args.output
    if cache_file is not None:
        print("Caching to %s" % cache_file)

    if args.csv:
        if cache_file is None:
            print(esdb.to_dataframe().to_csv())
        else:
            esdb.to_dataframe().to_csv(cache_file)
    else:
        esdb.save_cache(cache_file)



"""Connect to various outputs made available by HPSS
"""

import re
import copy
import datetime
from tokio.connectors.common import SubprocessOutputDict

REX_HEADING_LINE = re.compile(r"^[= ]+$")
REX_EMPTY_LINE = re.compile(r"^\s*$")
REX_TIMEDELTA = re.compile(r"^(\d+)-(\d+):(\d+):(\d+)$")

FLOAT_KEYS = set([
    'io_gb',
    'write_gb',
    'read_gb',
    'copy_gb',
    'mig (gb)',
    'purge(gb)',
    'lock%'
])
INT_KEYS = set([
    'users',
    'ops',
    'w_ops',
    'r_ops',
    'c_ops',
    'migfiles',
    'purfiles',
    'count',
    'cleans',
    'locks',
    'mounts',
])
DELTIM_KEYS = set([
    'migtime',
    'purgetime',
    'availtime',
    'locktime',
    'mounttime',
])

REKEY_TABLES = {
    'io totals by client application': 'client',
    'io totals by client host': 'host',
    'io totals by hpss client gateway (ui) host': 'host',
#   'largest users': 'user', # degenerate users will occur if one user uses multiple client apps
    'migration purge report': 'sc',
#   'tape drive report': 'drivetyp',
}

class HpssDailyReport(SubprocessOutputDict):
    """Representation for the daily report that HPSS can generate
    """
    def __init__(self, *args, **kwargs):
        super(HpssDailyReport, self).__init__(*args, **kwargs)
        self.date = None
        self.load()

    def load_str(self, input_str):
        """Parse the HPSS daily report text
        """
        lines = input_str.splitlines()
        num_lines = len(lines)
        start_line = 0

        # Look for the header for the whole report to get the report date
        for start_line, line in enumerate(lines):
            if line.startswith("HPSS Report for Date"):
                self.date = datetime.datetime.strptime(line.split()[-1], "%Y-%m-%d")
                break
        if not self.date:
            raise IndexError("No report date found")

        # Try to find tables encoded in the remainder of the report
        while start_line < num_lines:
            parsed_table, finish_line = _parse_section(lines, start_line)
            if finish_line != start_line and 'records' in parsed_table:
                if parsed_table['system'] not in self:
                    self.__setitem__(parsed_table['system'], {})

                # convert a list of records into a dict of indices
                if parsed_table['title'] in REKEY_TABLES:
                    parsed_table = _rekey_table(parsed_table,
                                                key=REKEY_TABLES[parsed_table['title']])

                self[parsed_table['system']][parsed_table['title']] = parsed_table['records']
            start_line += 1

def _parse_section(lines, start_line=0):
    """Parse a single table of the HPSS daily report

    Converts a table from the HPSS daily report into a dictionary.  For example
    an example table may appear as::

        Archive : IO Totals by HPSS Client Gateway (UI) Host
        Host             Users      IO_GB       Ops
        ===============  =====  =========  ========
        heart               53   148740.6     27991
        dtn11                5    29538.6      1694
        Total               58   178279.2     29685
        HPSS ACCOUNTING:         224962.6

    which will return a dict of form::

        {
            "system": "archive",
            "title": "io totals by hpss client gateway (ui) host",
            "records": {
                "heart": {
                    "io_gb": "148740.6",
                    "ops": "27991",
                    "users": "53",
                },
                "dtn11": {
                    "io_gb": "29538.6",
                    "ops": "1694",
                    "users": "5",
                },
                "total": {
                    "io_gb": "178279.2",
                    "ops": "29685",
                    "users": "58",
                }
            ]
        }

    This function is robust to invalid data, and any lines that do not appear to
    be a valid table will be treated as the end of the table.

    Args:
        lines (list of str): Text of the HPSS report
        start_line (int): Index of ``lines`` defined such that

          * ``lines[start_line]`` is the table title
          * ``lines[start_line + 1]`` is the table heading row
          * ``lines[start_line + 2]`` is the line separating the table heading and
            the first row of data
          * ``lines[start_line + 3:]`` are the rows of the table

    Returns:
        tuple:
          Tuple of (dict, int) where

          * dict contains the parsed contents of the table
          * int is the index of the last line of the table + 1
    """
    results = {}

    # Skip any initial whitespace
    num_lines = len(lines)
    while start_line < num_lines and REX_EMPTY_LINE.match(lines[start_line]):
        start_line += 1

    # Did we skip past the end of the input data?
    if start_line >= num_lines:
        return results, start_line

    # Parse table title (if available).  This can pick up times (0:00:00) so do
    # not treat system, title as legitimate values until we also identify the
    # line below column headings.
    if ':' not in lines[start_line]:
        return results, start_line
    else:
        system, title = lines[start_line].split(':', 1)

    # Determine column delimiters
    separator_line = lines[start_line + 2]
    col_extents = _find_columns(separator_line)
    if len(col_extents) == 0:
        return results, start_line

    # At this point, we are reasonably confident we have found a table.
    # Populate results so that this function returns some indicator of
    # success.
    results['system'] = system.strip().lower()
    results['title'] = title.strip().lower()

    # Determine column headers
    heading_line = lines[start_line + 1]
    headings = []
    for start_pos, str_len in col_extents:
        headings.append(heading_line[start_pos:start_pos + str_len].strip())

    records = []
    index = 0
    for index, line in enumerate(lines[start_line + 3:]):
        # check for end of record (empty line)
        if REX_EMPTY_LINE.match(line):
            # an empty line denotes end of table
            break
        elif len(line) < (col_extents[-1][0] + col_extents[-1][1] - 1):
            # line is malformed; this happens for table summaries
            break

        record = {}
        for heading_idx, (start_pos, str_len) in enumerate(col_extents):
            col_name = headings[heading_idx].lower()
            col_val = line[start_pos:start_pos + str_len].lower().strip()
            if col_name in FLOAT_KEYS:
                record[col_name] = float(col_val)
            elif col_name in INT_KEYS:
                record[col_name] = int(col_val)
            elif col_name in DELTIM_KEYS:
                record[col_name] = col_val
                record[col_name + "secs"] = _hpss_timedelta_to_secs(col_val)
            else:
                record[col_name] = col_val
        records.append(record)

    if records:
        results['records'] = records

    return (results, index + 1)

def _find_columns(line, sep="=", gap=' ', strict=False):
    """Determine the column start/end positions for a header line separator

    Takes a line separator such as the one denoted below:

        Host             Users      IO_GB
        ===============  =====  =========
        heart               53   148740.6

    and returns a tuple of (start index, end index) values that can be used to
    slice table rows into column entries.

    Args:
        line (str): Text comprised of separator characters and spaces that
            define the extents of columns
        sep (str): The character used to draw the column lines
        gap (str): The character separating ``sep`` characters
        strict (bool): If true, restrict column extents to only include sep
            characters and not the spaces that follow them.

    Returns:
        list of tuples:
    """

    columns = []

    # if line is not comprised exclusively of separators and gaps, it is not a
    # valid heading line
    if line.replace(sep, 'X').replace(gap, 'X').strip('X') != "":
        return columns

    if strict:
        col_start = None
    else:
        col_start = 0

    for index, char in enumerate(line):
        if strict:
            # col_start == None == looking for start of a new column
            if col_start is None and char == sep:
                col_start = index
            # if this is the end of an inter-column gap
            elif index > 0 and char == gap and line[index - 1] == sep:
                columns.append((col_start, index - col_start))
                col_start = None
        else:
            # if this is the end of an inter-column gap
            if index > 0 and char == gap and line[index - 1] == sep:
                columns.append((col_start, index - col_start))
                col_start = index

    if line and line[-1] == sep and col_start is not None:
        columns.append((col_start, len(line) - col_start))

    return columns

def _rekey_table(table, key):
    """Converts a list of records into a dict of records

    Converts a table of records as returned by _parse_section() of the form::

        {
            "records": [
                {
                    "host": "heart",
                    "io_gb": "148740.6",
                    "ops": "27991",
                    "users": "53",
                },
                ...
            ]
        }

    Into a table of key-value pairs the form::

        {
            "records": {
                "heart": {
                    "io_gb": "148740.6",
                    "ops": "27991",
                    "users": "53",
                },
                ...
            }
        }

    Does not handle degenerate keys when re-keying, so only some tables with a
    uniquely identifying key can be rekeyed.

    Args:
        table (dict): Output of the _parse_section() function
        key (str): Key to pull out of each element of table['records'] to use as
            the key for each record

    Returns:
        dict: Table with records expressed as key-value pairs instead of a list
    """
    new_table = copy.deepcopy(table)
    new_records = {}

    for record in new_table['records']:
        new_key = record.pop(key)
        if new_key in new_records:
            raise KeyError("Degenerate key %s=%s" % (key, new_key))
        new_records[new_key] = record

    new_table['records'] = new_records

    return new_table

def _hpss_timedelta_to_secs(timedelta_str):
    """Convert HPSS-encoded timedelta string into seconds

    Args:
        timedelta_str (str): String in form d-HH:MM:SS where d is the number of
            days, HH is hours, MM minutes, and SS seconds

    Returns:
        int: number of seconds represented by timedelta_str
    """

    match = REX_TIMEDELTA.match(timedelta_str)
    if match:
        seconds = int(match.group(1)) * 86400
        seconds += int(match.group(2)) * 3600
        seconds += int(match.group(3)) * 60
        seconds += int(match.group(4))
    else:
        seconds = -1

    return seconds
